"""Implementations for DisplayObject class.
"""

from typing import TYPE_CHECKING

from apysc._animation.animation_parallel_interface import \
    AnimationParallelInterface
from apysc._display.css_interface import CssInterface
from apysc._display.parent_interface import ParentInterface
from apysc._display.visible_interface import VisibleInterface
from apysc._display.x_interface import XInterface
from apysc._display.y_interface import YInterface
from apysc._event.custom_event_interface import CustomEventInterface
from apysc._event.mouse_event_interfaces import MouseEventInterfaces

if TYPE_CHECKING:
    from apysc._display.stage import Stage


class DisplayObject(
        XInterface, YInterface, ParentInterface, MouseEventInterfaces,
        VisibleInterface, CustomEventInterface, CssInterface,
        AnimationParallelInterface):
    """
    Display object (base) class for the common interfaces.

    References
    ----------
    - DisplayObject document
        - https://simon-ritchie.github.io/apysc/display_object.html
    """

    stage: 'Stage'

    def __init__(self, *, variable_name: str) -> None:
        """
        Display object (base) class for the common interfaces.

        Parameters
        ----------
        variable_name : str
            Variable name of this instance. This will be used to
            js expression.

        References
        ----------
        - DisplayObject document
            - https://simon-ritchie.github.io/apysc/display_object.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=DisplayObject):
            from apysc._validation import string_validation
            stage: ap.Stage = ap.get_stage()
            self.stage: ap.Stage = stage
            self._variable_name = variable_name
            string_validation.validate_not_empty_string(string=variable_name)

    def _set_overflow_visible_setting(self) -> None:
        """
        Set the `visible` value to the `overflow` CSS property.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._set_overflow_visible_setting, locals_=locals(),
                module_name=__name__, class_=DisplayObject):
            self.set_css(name='overflow', value='visible')



"""Type related common implementations.

Mainly following interface is defined:

- is_number
    - Get a boolean value whether specified value is Number value.
- is_float_or_number
    - Get a boolean value whether specified value is float or Nuber
    value.
- is_bool
    - Get a boolean value whether specified value is bool or Boolean
    value.
- is_same_class_instance
    - Get a boolean value whether specified class and instance's class
    are same or not.
- is_immutable_type
    - Get a boolean value whether specified value is immutable
    type or not.
"""

from typing import Any
from typing import Tuple
from typing import Type


def is_number(*, value: Any) -> bool:
    """
    Get a boolean value whether specified value is Number value.

    Parameters
    ----------
    value : *
        Any value to check.

    Returns
    -------
    result : bool
        If Number value is specified, True will be returned.
    """
    import apysc as ap
    if isinstance(value, ap.Number):
        return True
    return False


def is_float_or_number(*, value: Any) -> bool:
    """
    Get a boolean value whether specified value is float or Nuber
    value.

    Parameters
    ----------
    value : *
        Any value to check.

    Returns
    -------
    result : bool
        If float or Number value is specified, True will be returned.
    """
    if isinstance(value, float):
        return True
    if is_number(value=value):
        return True
    return False


def is_bool(*, value: Any) -> bool:
    """
    Get a boolean value whether specified value is bool or Boolean
    value.

    Parameters
    ----------
    value : *
        Any value to check.

    Returns
    -------
    result : bool
        If bool or Boolean value is specified, True will be returned.
    """
    if is_same_class_instance(class_=bool, instance=value):
        return True
    import apysc as ap
    if isinstance(value, ap.Boolean):
        return True
    return False


def is_same_class_instance(*, class_: Type, instance: Any) -> bool:
    """
    Get a boolean value whether specified class and instance's class
    are same or not.

    Notes
    -----
    If instance is subclass of `cls` argument, differ from `isinstace`,
    then False will be returned.

    Parameters
    ----------
    class_ : Type
        Expected class.
    instance : *
        Intance to check it's class.

    Returns
    -------
    result : bool
        If a specified class and instance's class are same, then True
        will be set.
    """
    instance_type: Type = type(instance)  # type: ignore
    if instance_type == class_:
        return True
    return False


def is_immutable_type(value: Any) -> bool:
    """
    Get a boolean value whether specified value is immutable
    type or not.

    Notes
    -----
    apysc's value types, such as the `Int`, are checked
    as immutable since these js types are immutable.

    Parameters
    ----------
    value : Any
        Target value to check.

    Returns
    -------
    result : bool
        If a specified value is immutable, then True
        will be set.
    """
    import apysc as ap
    immutable_types: Tuple = (
        int, float, bool, str, complex, tuple, range, bytes,
        ap.Int, ap.Number, ap.String, ap.Boolean,
    )
    if isinstance(value, immutable_types):
        return True
    return False



"""Class implementation for the animation_reset interface.
"""

from apysc._type.variable_name_interface import VariableNameInterface


class AnimationResetInterface(VariableNameInterface):

    def animation_reset(self) -> None:
        """
        Stop the all animations and reset.

        References
        ----------
        - animation_reset interface document
            - https://simon-ritchie.github.io/apysc/animation_reset.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.animation_reset, locals_=locals(),
                module_name=__name__, class_=AnimationResetInterface):
            expression: str = (
                f'{self.variable_name}.timeline().stop();'
            )
            ap.append_js_expression(expression=expression)



"""Class implementation for the animation_finish interface.
"""

from apysc._type.variable_name_interface import VariableNameInterface


class AnimationFinishInterface(VariableNameInterface):

    def animation_finish(self) -> None:
        """
        Finish the all animations (set the animation last value to each
        attribute).

        References
        ----------
        - animation_finish interface document
            - https://simon-ritchie.github.io/apysc/animation_finish.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.animation_finish, locals_=locals(),
                module_name=__name__, class_=AnimationFinishInterface):
            expression: str = (
                f'{self.variable_name}.timeline().finish();'
            )
            ap.append_js_expression(expression=expression)



"""Class implementation for fill color interface.
"""

from typing import Dict
from typing import Union

from apysc._animation.animation_fill_color_interface import \
    AnimationFillColorInterface
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.revert_interface import RevertInterface
from apysc._type.string import String


class FillColorInterface(
        AnimationFillColorInterface, RevertInterface, AttrLinkingInterface):

    _fill_color: String

    @property
    def fill_color(self) -> String:
        """
        Get this instance's fill color.

        Returns
        -------
        fill_color : String
            Current fill color (hexadecimal string, e.g., '#00aaff').
            If not be set, None will be returned.

        References
        ----------
        - Graphics fill_color interface document
            - https://simon-ritchie.github.io/apysc/graphics_fill_color.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> rectangle.fill_color = ap.String('#f0a')
        >>> rectangle.fill_color
        String('#ff00aa')
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='fill_color', locals_=locals(),
                module_name=__name__, class_=FillColorInterface):
            from apysc._type import value_util
            self._initialize_fill_color_if_not_initialized()
            fill_color: ap.String = value_util.get_copy(value=self._fill_color)
            return fill_color

    @fill_color.setter
    def fill_color(self, value: String) -> None:
        """
        Update this instance's fill color.

        Parameters
        ----------
        value : String
            Fill color to set.

        References
        ----------
        - Graphics fill_color interface document
            - https://simon-ritchie.github.io/apysc/graphics_fill_color.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='fill_color', locals_=locals(),
                module_name=__name__, class_=FillColorInterface):
            self._update_fill_color_and_skip_appending_exp(value=value)
            self._append_fill_color_update_expression()

            self._append_applying_new_attr_val_exp(
                new_attr=self._fill_color, attr_name='fill_color')
            self._append_attr_to_linking_stack(
                attr=self._fill_color, attr_name='fill_color')

    def _append_fill_color_update_expression(self) -> None:
        """
        Append fill color updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_fill_color_update_expression,
                locals_=locals(),
                module_name=__name__, class_=FillColorInterface):
            expression: str = (
                f'{self.variable_name}.fill("{self.fill_color}");'
            )
            ap.append_js_expression(expression=expression)

    def _set_initial_fill_color_if_not_blank(
            self, *, fill_color: Union[str, String]) -> None:
        """
        Set initial fill color value if specified value is not
        blank string.

        Parameters
        ----------
        fill_color : str or String
            Fill color (hexadecimal string, e.g., '#00aaff').
        """
        self._initialize_fill_color_if_not_initialized()
        if fill_color == '':
            return
        if isinstance(fill_color, str):
            fill_color = String(fill_color)
        self._update_fill_color_and_skip_appending_exp(value=fill_color)

    def _update_fill_color_and_skip_appending_exp(
            self, *, value: String) -> None:
        """
        Update fill color and skip appending expression.

        Parameters
        ----------
        value : String
            Fill color to set.
        """
        from apysc._color import color_util
        value = color_util.complement_hex_color(
            hex_color_code=value)
        self._initialize_fill_color_if_not_initialized()
        self._fill_color.value = value

    def _initialize_fill_color_if_not_initialized(self) -> None:
        """
        Initialize fill_color attribute if that value is not
        instantiated yet.
        """
        if hasattr(self, '_fill_color'):
            return
        self._fill_color = String('')

    _fill_color_snapshots: Dict[str, str]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_fill_color_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_fill_color_snapshots',
            value=self._fill_color._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._fill_color._value = self._fill_color_snapshots[snapshot_name]



"""Class implementations for the animation_move interface.
"""

from typing import Union

from apysc._animation.animation_interface_base import AnimationInterfaceBase
from apysc._animation.animation_move import AnimationMove
from apysc._animation.easing import Easing
from apysc._type.int import Int


class AnimationMoveInterface(AnimationInterfaceBase):

    def animation_move(
            self,
            x: Union[int, Int],
            y: Union[int, Int],
            *,
            duration: Union[int, Int] = 3000,
            delay: Union[int, Int] = 0,
            easing: Easing = Easing.LINEAR) -> AnimationMove:
        """
        Set the x and y coordinates animation settings.

        Notes
        -----
        To start this animation, you need to call the `start` method of
        the returned instance.

        Parameters
        ----------
        x : int or Int
            Destination of the x-coordinate.
        y : int or Int
            Destination of the y-coordinate.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Returns
        -------
        animation_move : AnimationMove
            Created animation setting instance.

        References
        ----------
        - animation_move interface document
            - https://simon-ritchie.github.io/apysc/animation_move.html
        - Animation interfaces duration setting document
            - https://simon-ritchie.github.io/apysc/animation_duration.html
        - Animation interfaces delay setting document
            - https://simon-ritchie.github.io/apysc/animation_delay.html
        - Each animation interface return value document
            - https://bit.ly/2XOoa8w
        - Sequential animation setting document
            - https://simon-ritchie.github.io/apysc/sequential_animation.html
        - animation_parallel interface document
            - https://simon-ritchie.github.io/apysc/animation_parallel.html
        - Easing enum document
            - https://simon-ritchie.github.io/apysc/easing_enum.html
        """
        animation_move: AnimationMove = AnimationMove(
            target=self,
            x=x,
            y=y,
            duration=duration,
            delay=delay,
            easing=easing)
        return animation_move



"""Class implementation for line cap interface.
"""

from typing import Dict
from typing import Union

from apysc._display.line_caps import LineCaps
from apysc._type.revert_interface import RevertInterface
from apysc._type.string import String
from apysc._type.variable_name_interface import VariableNameInterface


class LineCapInterface(VariableNameInterface, RevertInterface):

    _line_cap: String

    def _initialize_line_cap_if_not_initialized(self) -> None:
        """
        Inilialize _line_cap attribute if it is not
        initialized yet.
        """
        if hasattr(self, '_line_cap'):
            return
        self._line_cap = String(LineCaps.BUTT.value)

    @property
    def line_cap(self) -> Union[String, LineCaps]:
        """
        Get this instance's line cap style setting.

        Returns
        -------
        line_cap : String
            Line cap style setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='line_cap', locals_=locals(),
                module_name=__name__, class_=LineCapInterface):
            self._initialize_line_cap_if_not_initialized()
            return self._line_cap._copy()

    @line_cap.setter
    def line_cap(self, value: Union[String, LineCaps]) -> None:
        """
        Set line cap style setting.

        Parameters
        ----------
        value : String or LineCaps
            Line cap style setting to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='line_cap', locals_=locals(),
                module_name=__name__, class_=LineCapInterface):
            self._update_line_cap_and_skip_appending_exp(value=value)
            self._append_line_cap_update_expression()

    def _append_line_cap_update_expression(self) -> None:
        """
        Append line cap updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_line_cap_update_expression,
                locals_=locals(),
                module_name=__name__, class_=LineCapInterface):
            from apysc._type import value_util
            cap_name: str = value_util.get_value_str_for_expression(
                value=self._line_cap)
            expression: str = (
                f'{self.variable_name}.attr({{"stroke-linecap": {cap_name}}});'
            )
            ap.append_js_expression(expression=expression)

    def _update_line_cap_and_skip_appending_exp(
            self, *, value: Union[String, LineCaps]) -> None:
        """
        Update line cap and skip appending expression.

        Parameters
        ----------
        value : String or LineCaps
            Line cap style setting to set.
        """
        from apysc._validation.display_validation import validate_line_cap
        if not isinstance(value, (String, LineCaps)):
            raise TypeError(
                'Not supported line_cap type specified: '
                f'{type(value)}'
                '\nAcceptable ones are: String or LineCaps.')
        validate_line_cap(cap=value)
        if isinstance(value, String):
            self._line_cap = value._copy()
        else:
            self._line_cap = String(value.value)

    _line_cap_snapshots: Dict[str, str]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_line_cap_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_line_cap_snapshots',
            value=self._line_cap._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._line_cap._value = self._line_cap_snapshots[snapshot_name]



"""Path data class implementation for the svg's
`continual 3D bezier curve` (S).
"""

from typing import Union

from apysc._geom.path_control_x_interface import PathControlXInterface
from apysc._geom.path_control_y_interface import PathControlYInterface
from apysc._geom.path_data_base import PathDataBase
from apysc._geom.path_dest_x_interface import PathDestXInterface
from apysc._geom.path_dest_y_interface import PathDestYInterface
from apysc._type.boolean import Boolean
from apysc._type.int import Int
from apysc._type.string import String


class PathBezier3DContinual(
        PathDataBase, PathControlXInterface, PathControlYInterface,
        PathDestXInterface, PathDestYInterface):
    """
    Path data class for the svg's `continual 3D bezier curve` (S).
    """

    def __init__(
            self,
            control_x: Union[int, Int],
            control_y: Union[int, Int],
            dest_x: Union[int, Int],
            dest_y: Union[int, Int],
            *,
            relative: Union[bool, Boolean] = False) -> None:
        """
        Path data class for the svg's `continual 3D bezier curve` (S).

        Parameters
        ----------
        control_x : int or Int
            X-coordinate of the bezier's control point.
        control_y : int or Int
            Y-coordinate of the bezier's control point.
        dest_x : int or Int
            X-coordinate of the destination point.
        dest_y : int or Int
            Y-coordinate of the destination point.
        relative : bool or Boolean, default False
            The boolean value indicating whether the path
            coordinates are relative or not (absolute).
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=PathBezier3DContinual):
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_int_from_builtin_val
            from apysc._geom.path_label import PathLabel
            super(PathBezier3DContinual, self).__init__(
                path_label=PathLabel.BEZIER_3D_CONTINUAL,
                relative=relative)
            self.control_x = get_copied_int_from_builtin_val(
                integer=control_x)
            self.control_y = get_copied_int_from_builtin_val(
                integer=control_y)
            self.dest_x = get_copied_int_from_builtin_val(integer=dest_x)
            self.dest_y = get_copied_int_from_builtin_val(integer=dest_y)

    def _get_svg_str(self) -> str:
        """
        Get a path's SVG string created with the current setting.

        Returns
        -------
        svg_str : str
            A path's SVG string created with the current setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._get_svg_str, locals_=locals(),
                module_name=__name__, class_=PathBezier3DContinual):
            from apysc._type.value_util import get_value_str_for_expression
            svg_char: String = self._get_svg_char()
            svg_char_str: str = get_value_str_for_expression(
                value=svg_char)
            control_x_str: str = get_value_str_for_expression(
                value=self._control_x)
            control_y_str: str = get_value_str_for_expression(
                value=self._control_y)
            dest_x_str: str = get_value_str_for_expression(value=self._dest_x)
            dest_y_str: str = get_value_str_for_expression(value=self._dest_y)
            svg_str: str = (
                f'{svg_char_str} + String({control_x_str}) '
                f'+ " " + String({control_y_str}) '
                f'+ " " + String({dest_x_str}) '
                f'+ " " + String({dest_y_str})'
            )
            return svg_str

    def update_path_data(
            self,
            control_x: Union[int, Int],
            control_y: Union[int, Int],
            dest_x: Union[int, Int],
            dest_y: Union[int, Int],
            relative: Union[bool, Boolean]) -> None:
        """
        Update the path's data settings.

        Parameters
        ----------
        control_x : int or Int
            X-coordinate of the bezier's control point.
        control_y : int or Int
            Y-coordinate of the bezier's control point.
        dest_x : int or Int
            X-coordinate of the destination point.
        dest_y : int or Int
            Y-coordinate of the destination point.
        relative : bool or Boolean, default False
            The boolean value indicating whether the path
            coordinates are relative or not (absolute).
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.update_path_data, locals_=locals(),
                module_name=__name__, class_=PathBezier3DContinual):
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_boolean_from_builtin_val
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_int_from_builtin_val
            self.control_x = get_copied_int_from_builtin_val(
                integer=control_x)
            self.control_y = get_copied_int_from_builtin_val(
                integer=control_y)
            self.dest_x = get_copied_int_from_builtin_val(integer=dest_x)
            self.dest_y = get_copied_int_from_builtin_val(integer=dest_y)
            self.relative = get_copied_boolean_from_builtin_val(
                bool_val=relative)



"""Graphics class related expression implementations.
"""

from apysc._display.graphics import Graphics
from apysc._display.graphics_base import GraphicsBase


def append_fill_expression(
        *, graphics: Graphics, expression: str, indent_num: int) -> str:
    """
    Append fill expression to specified expression's string.

    Parameters
    ----------
    graphics : Graphics
        Target Graphics instance.
    expression : str
        Expression string to be appended fill expression.
    indent_num : int
        Indentation number.

    Returns
    -------
    expression : str
        After appended expression string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_fill_expression, locals_=locals(),
            module_name=__name__):
        from apysc._string import indent_util
        spaces: str = indent_util.make_spaces_for_html(indent_num=indent_num)
        if graphics.fill_color == '':
            expression += f'\n{spaces}fill: "none",'
            return expression
        expression += (
            f'\n{spaces}fill: {graphics.fill_color.variable_name},'
        )
        return expression


def append_fill_opacity_expression(
        *, graphics: Graphics, expression: str, indent_num: int) -> str:
    """
    Append fill opacity expression to specified expression's string.

    Parameters
    ----------
    graphics : Graphics
        Target Graphics instance.
    expression : str
        Expression string to be appended fill opacity expression.
    indent_num : int
        Indentation number.

    Returns
    -------
    expression : str
        After appended expression string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_fill_opacity_expression, locals_=locals(),
            module_name=__name__):
        from apysc._string import indent_util
        spaces: str = indent_util.make_spaces_for_html(indent_num=indent_num)
        expression += (
            f'\n{spaces}"fill-opacity": {graphics.fill_alpha.variable_name},'
        )
        return expression


def append_x_expression(
        *, graphic: GraphicsBase, expression: str, indent_num: int) -> str:
    """
    Append x position expression to specified expression's string.

    Parameters
    ----------
    graphic : GraphicsBase
        Target graphic instance, for example, Rectangle.
    expression : str
        Expression string to be appended x position expression.
    indent_num : int
        Indentation number.

    Returns
    -------
    expression : str
        After appended expression string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_x_expression, locals_=locals(),
            module_name=__name__):
        from apysc._string import indent_util
        spaces: str = indent_util.make_spaces_for_html(indent_num=indent_num)
        expression += (
            f'\n{spaces}x: {graphic.x.variable_name},'
        )
        return expression


def append_y_expression(
        *, graphic: GraphicsBase, expression: str, indent_num: int) -> str:
    """
    Append y position expression to specified expression's string.

    Parameters
    ----------
    graphic : GraphicsBase
        Target graphic instance, for example, Rectangle.
    expression : str
        Expression string to be appended y position expression.
    indent_num : int
        Indentation number.

    Returns
    -------
    expression : str
        After appended expression string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_y_expression, locals_=locals(),
            module_name=__name__):
        from apysc._string import indent_util
        spaces: str = indent_util.make_spaces_for_html(indent_num=indent_num)
        expression += (
            f'\n{spaces}y: {graphic.y.variable_name},'
        )
        return expression


def append_stroke_expression(
        *, graphics: Graphics, expression: str, indent_num: int) -> str:
    """
    Append stroke expression to specified expression's string.

    Parameters
    ----------
    graphics : Graphics
        Target Graphics instance.
    expression : str
        Expression string to be appended stroke expression.
    indent_num : int
        Indentation number.

    Returns
    -------
    expression : str
        After appended expression string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_stroke_expression, locals_=locals(),
            module_name=__name__):
        from apysc._string import indent_util
        if graphics.line_color == '':
            return expression
        spaces: str = indent_util.make_spaces_for_html(indent_num=indent_num)
        expression += (
            f'\n{spaces}stroke: {graphics.line_color.variable_name},'
        )
        return expression


def append_stroke_width_expression(
        *, graphics: Graphics, expression: str, indent_num: int) -> str:
    """
    Append stroke width expression to specified expression's string.

    Parameters
    ----------
    graphics : Graphics
        Target Graphics instance.
    expression : str
        Expression string to be appended stroke width expression.
    indent_num : int
        Indentation number.

    Returns
    -------
    expression : str
        After appended expression string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_stroke_width_expression, locals_=locals(),
            module_name=__name__):
        from apysc._string import indent_util
        spaces: str = indent_util.make_spaces_for_html(indent_num=indent_num)
        expression += (
            f'\n{spaces}"stroke-width": '
            f'{graphics.line_thickness.variable_name},'
        )
        return expression


def append_stroke_opacity_expression(
        *, graphics: Graphics, expression: str, indent_num: int) -> str:
    """
    Append stroke opacity expression to specified expression's string.

    Parameters
    ----------
    graphics : Graphics
        Target Graphics instance.
    expression : str
        Expression string to be appended stroke opacity expression.
    indent_num : int
        Indentation number.

    Returns
    -------
    expression : str
        After appended expression string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_stroke_opacity_expression, locals_=locals(),
            module_name=__name__):
        from apysc._string import indent_util
        spaces: str = indent_util.make_spaces_for_html(indent_num=indent_num)
        expression += (
            f'\n{spaces}"stroke-opacity": {graphics.line_alpha.variable_name},'
        )
        return expression


def append_stroke_linecap_expression(
        *, graphics: Graphics, expression: str, indent_num: int) -> str:
    """
    Append stroke linecap expression to specified expression's string.

    Parameters
    ----------
    graphics : Graphics
        Target Graphics instance.
    expression : str
        Expression string to be appended stroke linecap expression.
    indent_num : int
        Indentation number.

    Returns
    -------
    expression : str
        After appended expression string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_stroke_linecap_expression, locals_=locals(),
            module_name=__name__):
        from apysc._string import indent_util
        if graphics.line_cap == ap.LineCaps.BUTT.value:
            return expression
        spaces: str = indent_util.make_spaces_for_html(indent_num=indent_num)
        expression += (
            f'\n{spaces}"stroke-linecap": {graphics.line_cap.variable_name},'
        )
        return expression


def append_stroke_linejoin_expression(
        *, graphics: Graphics, expression: str, indent_num: int) -> str:
    """
    Append stroke linejoin expression to specified expression's string.

    Parameters
    ----------
    graphics : Graphics
        Target Graphics instance.
    expression : str
        Expression string to be appended stroke linejoin expression.
    indent_num : int
        Indentation number.

    Returns
    -------
    expression : str
        After appended expression string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=append_stroke_linejoin_expression, locals_=locals(),
            module_name=__name__):
        from apysc._string import indent_util
        if graphics.line_joints == ap.LineJoints.MITER.value:
            return expression
        spaces: str = indent_util.make_spaces_for_html(indent_num=indent_num)
        expression += (
            f'\n{spaces}"stroke-linejoin": '
            f'{graphics.line_joints.variable_name},'
        )
        return expression



"""Color related implementations.
"""

from typing import Any
from typing import TypeVar

from apysc._type.string import String

StrOrString = TypeVar('StrOrString', str, String)


def complement_hex_color(
        *, hex_color_code: StrOrString) -> StrOrString:
    """
    Complement hex color for convenience, for instance, add # prefix or
    three digits to six digits, upper case to lower case etc.

    Parameters
    ----------
    hex_color_code : str or String
        Hexadecimal color code. e.g., 'ff0000', '#666', '#0'

    Returns
    -------
    complemented_hex_color_code : str or String
        Result hex color code. e.g., '#ff0000', '#666666, '#000000'
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=complement_hex_color, locals_=locals(),
            module_name=__name__):
        from apysc._html import html_util
        from apysc._validation import color_validation
        hex_color_code = html_util.remove_first_selector_symbol_char(
            str_val=hex_color_code)

        color_validation.validate_hex_color_code_format(
            hex_color_code=hex_color_code)
        if isinstance(hex_color_code, ap.String):
            value_: str = hex_color_code._value
        else:
            value_ = hex_color_code
        char_len = len(value_)
        if char_len == 1:
            value_ = _fill_one_digit_hex_color_code(
                hex_color_code=value_)
        elif char_len == 3:
            value_ = _fill_three_digit_hex_color_code(
                hex_color_code=value_)
        value_ = value_.lower()
        value_ = f'#{value_}'

        if isinstance(hex_color_code, ap.String):
            hex_color_code._value = value_
            _append_complement_hex_color_expression(
                hex_color_code=hex_color_code)
        else:
            hex_color_code = value_  # type: ignore
        return hex_color_code  # type: ignore


def _append_complement_hex_color_expression(
        *, hex_color_code: Any) -> None:
    """
    Append complement_hex_color function's expression.

    Parameters
    ----------
    hex_color_code : String
        Complemented hex color code string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=_append_complement_hex_color_expression,
            locals_=locals(),
            module_name=__name__):
        import apysc as ap
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        index_name: str = expression_variables_util.get_next_variable_name(
            type_name=var_names.INDEX)
        hex_color_code_: ap.String = hex_color_code
        var_name: str = hex_color_code_.variable_name
        expression: str = (
            f'var str_length = {var_name}.length;'
            '\nif (str_length === 1) {'
            f'\n  {var_name} = "00000" + {var_name};'
            '\n}else if (str_length === 3) {'
            f'\n  var {var_name}_ = "";'
            f'\n  for (var {index_name} = 0; {index_name} < '
            f'{var_name}.length; '
            f'{index_name}++) {{'
            f'\n    {var_name}_ += {var_name}[{index_name}].repeat(2);'
            '\n  }'
            f'\n  {var_name} = {var_name}_;'
            '\n}'
            f'\n{var_name} = "#" + {var_name};')
        ap.append_js_expression(expression=expression)


def _fill_three_digit_hex_color_code(*, hex_color_code: str) -> str:
    """
    Fill 3 digits hexadecimal color code until it becomes 6 digits.

    Parameters
    ----------
    hex_color_code : str
        One digit hexadecimal color code (not including '#').
        e.g., 'aaa', 'fff'

    Returns
    -------
    filled_color_code : str
        Result color code. e.g., 'aaaaaa', 'ffffff'
    """
    filled_color_code: str = ''
    for char in hex_color_code:
        filled_color_code += char * 2
    return filled_color_code


def _fill_one_digit_hex_color_code(*, hex_color_code: str) -> str:
    """
    Fill 1 digit hexadecimal color code until it becomes 6 digits.

    Parameters
    ----------
    hex_color_code : str
        One digit hexadecimal color code (not including '#').
        e.g., 'a', '0'

    Returns
    -------
    filled_color_code : str
        Result color code. e.g., '00000a', '000000'
    """
    filled_color_code: str = hex_color_code.zfill(6)
    return filled_color_code



"""Base class implementation for the path data.
"""

from abc import ABC
from abc import abstractmethod
from typing import Union

from apysc._branch._else import Else
from apysc._branch._if import If
from apysc._geom.path_label import PathLabel
from apysc._geom.relative_interface import RelativeInterface
from apysc._type.boolean import Boolean
from apysc._type.string import String


class PathDataBase(RelativeInterface, ABC):
    """
    Base class for the path data.
    """

    _path_label: PathLabel

    def __init__(
            self, *, path_label: PathLabel,
            relative: Union[bool, Boolean]) -> None:
        """
        Base class for the path data.

        Parameters
        ----------
        path_label : PathLabel
            Target (svg's) path label.
        relative : bool or Boolean
            The boolean value indicating whether the path
            coordinates are relative or not (absolute).
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=PathDataBase):
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_boolean_from_builtin_val
            self._path_label = path_label
            self.relative = get_copied_boolean_from_builtin_val(
                bool_val=relative)

    def _get_svg_char(self) -> String:
        """
        Get a SVG character (e.g., 'M' or 'm') from the
        current setting.

        Returns
        -------
        svg_char : String
            Target SVG character.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._get_svg_char, locals_=locals(),
                module_name=__name__, class_=PathDataBase):
            svg_char_: str = self._path_label.value
            if self._relative._value:
                svg_char: String = String(svg_char_.lower())
            else:
                svg_char = String(svg_char_)
            with If(self._relative, locals_=locals()):
                svg_char.value = svg_char_.lower()
            with Else(locals_=locals()):
                svg_char.value = svg_char_
            return svg_char

    @abstractmethod
    def _get_svg_str(self) -> str:
        """
        Get a path's SVG string created with the current setting.
        """



"""Class implementation for the animation event.
"""

from typing import Generic
from typing import TypeVar

from apysc._animation import animation_base
from apysc._event.event import Event
from apysc._type.variable_name_interface import VariableNameInterface

_T = TypeVar('_T', bound=VariableNameInterface)


class AnimationEvent(Event, Generic[_T]):
    """
    Animation event class.
    """

    _this: 'animation_base.AnimationBase[_T]'

    def __init__(
            self, *,
            this: 'animation_base.AnimationBase[_T]') -> None:
        """
        Animation event class.

        Parameters
        ----------
        this : AnimationBase
            Animation setting instance.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=AnimationEvent):
            from apysc._expression import var_names
            super(AnimationEvent, self).__init__(
                this=this, type_name=var_names.ANIMATION_EVENT)

    @property
    def this(self) -> 'animation_base.AnimationBase[_T]':
        """
        Get a animation setting instance that linstening this event.

        Returns
        -------
        this : AnimationBase
            Instance that linstening this event.
        """
        return self._this



"""Class implementation for visible interface.
"""

from typing import Dict

from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.boolean import Boolean
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class VisibleInterface(
        VariableNameInterface, RevertInterface, AttrLinkingInterface):

    _visible: Boolean

    def _initialize_visible_if_not_initialized(self) -> None:
        """
        Initialize _visible attribute if it hasn't been initialized yet.
        """
        if hasattr(self, '_visible'):
            return
        self._visible = Boolean(True)

        self._append_visible_attr_linking_setting()

    def _append_visible_attr_linking_setting(self) -> None:
        """
        Append a visible attribute linking setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_visible_attr_linking_setting,
                locals_=locals(),
                module_name=__name__, class_=VisibleInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._visible, attr_name='visible')
            self._append_attr_to_linking_stack(
                attr=self._visible, attr_name='visible')

    @property
    def visible(self) -> Boolean:
        """
        Get a visibility of this instance.

        Returns
        -------
        result : Boolean
            If this instance is visible, True will be returned.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='visible', locals_=locals(),
                module_name=__name__, class_=VisibleInterface):
            from apysc._type import value_util
            self._initialize_visible_if_not_initialized()
            return value_util.get_copy(value=self._visible)

    @visible.setter
    def visible(self, value: Boolean) -> None:
        """
        Update visibility of this instance.

        Parameters
        ----------
        value : Boolean
            Boolean value to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='visible', locals_=locals(),
                module_name=__name__, class_=VisibleInterface):
            from apysc._validation import bool_validation
            bool_validation.validate_bool(value=value)
            if isinstance(value, bool):
                value = ap.Boolean(value)
            self._visible = value
            self._append_visible_update_expression()

            self._append_visible_attr_linking_setting()

    def _append_visible_update_expression(self) -> None:
        """
        Append visible property updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_visible_update_expression,
                locals_=locals(),
                module_name=__name__, class_=VisibleInterface):
            expression: str = (
                f'if ({self._visible.variable_name}) {{'
                f'\n  {self.variable_name}.show();'
                '\n}else {'
                f'\n  {self.variable_name}.hide();'
                '\n}'
            )
            ap.append_js_expression(expression=expression)

    _visible_snapshots: Dict[str, bool]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_visible_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_visible_snapshots',
            value=self._visible._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value is snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._visible._value = self._visible_snapshots[snapshot_name]



"""Helper module for the rotation interfaces.
"""

from typing import Union

from apysc._type.expression_string import ExpressionString
from apysc._type.int import Int


def get_coordinates_key_for_expression(
        *,
        x: Union[int, Int],
        y: Union[int, Int]) -> ExpressionString:
    """
    Get a key string for the expression from the x and y coordinates.

    Parameters
    ----------
    x : int or Int
        X-coordinate.
    y : int or Int
        Y-coordinate.

    Returns
    -------
    key_exp_str : ExpressionString
        Key expression string.
    """
    from apysc._type.variable_name_interface import VariableNameInterface
    if isinstance(x, VariableNameInterface):
        x_str: str = x.variable_name
    else:
        x_str = str(x)
    if isinstance(y, VariableNameInterface):
        y_str: str = y.variable_name
    else:
        y_str = str(y)
    key_exp_str: ExpressionString = ExpressionString(
        value=f'String({x_str}) + "_" + String({y_str})')
    return key_exp_str



"""Class implementation for the rotation_around_center_interface
interface.
"""

from typing import Dict

from apysc._animation.animation_rotation_around_center_interface import \
    AnimationRotationAroundCenterInterface
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class RotationAroundCenterInterface(
        AnimationRotationAroundCenterInterface, RevertInterface,
        AttrLinkingInterface):

    _rotation_around_center: Int

    def _initialize_rotation_around_center_if_not_initialized(self) -> None:
        """
        Initialize the `_rotation_around_center` attribute if if hasn't
        been initialized yet.
        """
        if hasattr(self, '_rotation_around_center'):
            return
        self._rotation_around_center = Int(0)

        self._append_rotation_around_center_attr_linking_setting()

    def _append_rotation_around_center_attr_linking_setting(self) -> None:
        """
        Append a rotation around center attribute linking setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.
                _append_rotation_around_center_attr_linking_setting,
                locals_=locals(),
                module_name=__name__, class_=RotationAroundCenterInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._rotation_around_center,
                attr_name='rotation_around_center')
            self._append_attr_to_linking_stack(
                attr=self._rotation_around_center,
                attr_name='rotation_around_center')

    @property
    def rotation_around_center(self) -> Int:
        """
        Get a rotation value around the center of this instance.

        Returns
        -------
        rotation_around_center : Int
            Rotation value around the center of this instance.

        References
        ----------
        - GraphicsBase rotation_around_center interface document
            - https://bit.ly/3AjeSPr
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='rotation_around_center', locals_=locals(),
                module_name=__name__, class_=RotationAroundCenterInterface):
            from apysc._type import value_util
            self._initialize_rotation_around_center_if_not_initialized()
            return value_util.get_copy(value=self._rotation_around_center)

    @rotation_around_center.setter
    def rotation_around_center(self, value: Int) -> None:
        """
        Update a rotation value around the center of this instance.

        Parameters
        ----------
        value : int or Int
            Rotation value around the center of this instance.

        References
        ----------
        - GraphicsBase rotation_around_center interface
            - https://bit.ly/3AjeSPr
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='rotation_around_center', locals_=locals(),
                module_name=__name__, class_=RotationAroundCenterInterface):
            from apysc._validation import number_validation
            self._initialize_rotation_around_center_if_not_initialized()
            number_validation.validate_integer(integer=value)
            if not isinstance(value, ap.Int):
                value = ap.Int(value)
            before_value: ap.Int = self._rotation_around_center
            self._rotation_around_center = value
            self._append_rotation_around_center_update_expression(
                before_value=before_value)

            self._append_rotation_around_center_attr_linking_setting()

    def _append_rotation_around_center_update_expression(
            self, *, before_value: Int) -> None:
        """
        Append the rotation around the center of this instance
        updating expression.

        Parameters
        ----------
        before_value : ap.Int
            Before updating value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_rotation_around_center_update_expression,  # noqa
                locals_=locals(),
                module_name=__name__, class_=RotationAroundCenterInterface):
            from apysc._type import value_util
            before_value_str: str = value_util.get_value_str_for_expression(
                value=before_value)
            after_value_str: str = value_util.get_value_str_for_expression(
                value=self._rotation_around_center)
            expression: str = (
                f'{self.variable_name}.rotate(-{before_value_str});'
                f'\n{self.variable_name}.rotate({after_value_str});'
                f'\n{before_value_str} = {after_value_str};'
            )
            ap.append_js_expression(expression=expression)

    _rotation_around_center_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_rotation_around_center_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_rotation_around_center_snapshots',
            value=int(self._rotation_around_center._value),
            snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._rotation_around_center._value = \
            self._rotation_around_center_snapshots[snapshot_name]



"""Class implementation for the flip_x interface.
"""

from typing import Dict

from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.boolean import Boolean
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class FlipXInterface(
        VariableNameInterface, RevertInterface, AttrLinkingInterface):

    _flip_x: Boolean

    def _initialize_flip_x_if_not_initialized(self) -> None:
        """
        Initialize the _flip_x attribute if it hasn't been
        initialized yet.
        """
        if hasattr(self, '_flip_x'):
            return
        self._flip_x = Boolean(False)

        self._append_flip_x_attr_linking_setting()

    def _append_flip_x_attr_linking_setting(self) -> None:
        """
        Append a flip-x attribute linking setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_flip_x_attr_linking_setting,
                locals_=locals(),
                module_name=__name__, class_=FlipXInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._flip_x, attr_name='flip_x')
            self._append_attr_to_linking_stack(
                attr=self._flip_x, attr_name='flip_x')

    @property
    def flip_x(self) -> Boolean:
        """
        Get a boolean value whether the x-axis is flipped or not.

        Returns
        -------
        flip_x : Boolean
            A boolean value whether the x-axis is flipped or not.

        References
        ----------
        - GraphicsBase flip_x and flip_y interfaces document
            - https://bit.ly/31F9lHv

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> polygon: ap.Polygon = sprite.graphics.draw_polygon(
        ...     points=[
        ...         ap.Point2D(x=0, y=0),
        ...         ap.Point2D(x=0, y=50),
        ...         ap.Point2D(x=50, y=25),
        ...     ])
        >>> polygon.flip_x = ap.Boolean(True)
        >>> polygon.flip_x
        Boolean(True)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='flip_x', locals_=locals(),
                module_name=__name__, class_=FlipXInterface):
            from apysc._type import value_util
            self._initialize_flip_x_if_not_initialized()
            return value_util.get_copy(value=self._flip_x)

    @flip_x.setter
    def flip_x(self, value: Boolean) -> None:
        """
        Update a x-axis flipping value.

        Parameters
        ----------
        value : Boolean
            Flipping value. If True, a x-axis will be flipped,
            otherwise it will be reset.

        References
        ----------
        - GraphicsBase flip_x and flip_y interfaces document
            - https://bit.ly/31F9lHv
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='flip_x', locals_=locals(),
                module_name=__name__, class_=FlipXInterface):
            self._initialize_flip_x_if_not_initialized()
            before_value: ap.Boolean = self._flip_x
            self._flip_x = value
            self._append_flip_x_update_expression(before_value=before_value)

            self._append_flip_x_attr_linking_setting()

    def _append_flip_x_update_expression(
            self, *, before_value: Boolean) -> None:
        """
        Append a x-axis flipping value updating expression.

        Parameters
        ----------
        before_value : Boolean
            Before updating flipping value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_flip_x_update_expression,
                locals_=locals(),
                module_name=__name__, class_=FlipXInterface):
            from apysc._display import flip_interface_helper
            self._initialize_flip_x_if_not_initialized()
            expression: str = flip_interface_helper.\
                make_flip_update_expression(
                    before_value=before_value, after_value=self._flip_x,
                    axis=flip_interface_helper.Axis.X,
                    interface_variable_name=self.variable_name)
            ap.append_js_expression(expression=expression)

    _flip_x_snapshots: Dict[str, bool]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_flip_x_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_flip_x_snapshots',
            value=self._flip_x._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._flip_x._value = self._flip_x_snapshots[snapshot_name]



"""Dot setting class implementation for line.
"""

from typing import Union

from apysc._type.dictionary import Dictionary
from apysc._type.int import Int


class LineDotSetting(Dictionary[str, Int]):
    """
    Dot setting class for a line.

    References
    ----------
    - Graphics line_style interface document
        - https://bit.ly/3zauILT
    """

    def __init__(self, dot_size: Union[int, Int]) -> None:
        """
        Dot setting class for a line.

        Parameters
        ----------
        dot_size : int or Int
            Dot size.

        References
        ----------
        - Graphics line_style interface document
            - https://bit.ly/3zauILT
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=LineDotSetting):
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_int_from_builtin_val
            from apysc._validation import number_validation
            number_validation.validate_nums_are_int_and_gt_zero(
                nums=[dot_size])
            dot_size_: ap.Int = get_copied_int_from_builtin_val(
                integer=dot_size)
            super(LineDotSetting, self).__init__({'dot_size': dot_size_})

    @property
    def dot_size(self) -> Int:
        """
        Get a dot size setting.

        Returns
        -------
        dot_size : Int
            Dot size setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='dot_size', locals_=locals(),
                module_name=__name__, class_=LineDotSetting):
            return self['dot_size']._copy()



"""If branch instruction implementations.
"""

from typing import Any
from typing import Dict
from typing import Optional

from apysc._branch.if_base import IfBase
from apysc._type.boolean import Boolean


class If(IfBase):
    """
    A class to append if branch instruction expression.

    References
    ----------
    - If document
        - https://simon-ritchie.github.io/apysc/if.html
    - Each branch instruction class scope variables reverting setting
        - https://bit.ly/3rkAuaT
    """

    def __init__(
            self,
            condition: Optional[Boolean],
            *,
            locals_: Optional[Dict[str, Any]] = None,
            globals_: Optional[Dict[str, Any]] = None) -> None:
        """
        A class to append if branch instruction expression.

        Parameters
        ----------
        condition : Boolean or None
            Boolean value to be used for judgment.
        locals_ : dict or None, default None
            Current scope's local variables. Set locals() value to
            this argument. If specified, all local scope
            VariableNameInterface variables (like Int, Sprite) will be
            reverted ad the end of If scope. This setting is useful
            when you don't want to update each variable by the
            implementation of the If scope.
        globals_ : dict or None, default None
            Current scope's global variables. Set golobals() value
            to this argument. This works the same way as the locals_
            argument.

        References
        ----------
        - If document
            - https://simon-ritchie.github.io/apysc/if.html
        - Each branch instruction class scope variables reverting setting
            - https://bit.ly/3rkAuaT

        Examples
        --------
        >>> import apysc as ap
        >>> int_val: ap.Int = ap.Int(10)
        >>> condition: ap.Boolean = int_val >= 10
        >>> with ap.If(condition):
        ...     ap.trace('Int value is greater than equal 10!')
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=If):
            super(If, self).__init__(
                condition=condition, locals_=locals_, globals_=globals_)

    def _append_enter_expression(self) -> None:
        """
        Append if branch instruction start expression.
        """
        import apysc as ap
        if self._condition is None:
            raise ValueError(
                'If expression\'s condition argument can\'t be set None.')
        expression: str = (
            f'if ({self._condition.variable_name}) {{'
        )
        ap.append_js_expression(expression=expression)

    def _set_last_scope(self) -> None:
        """
        Set expression last scope value.
        """
        from apysc._expression import last_scope
        from apysc._expression.last_scope import LastScope
        last_scope.set_last_scope(value=LastScope.IF)



"""Class implementation for fill alpha interface.
"""

from typing import Dict

from apysc._animation.animation_fill_alpha_interface import \
    AnimationFillAlphaInterface
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.number import Number
from apysc._type.number_value_interface import NumberValueInterface
from apysc._type.revert_interface import RevertInterface


class FillAlphaInterface(
        AnimationFillAlphaInterface, RevertInterface, AttrLinkingInterface):

    _fill_alpha: Number

    def _initialize_fill_alpha_if_not_initialized(self) -> None:
        """
        Initialize _fill_alpha attribute if it hasn't been initialized yet.
        """
        if hasattr(self, '_fill_alpha'):
            return
        self._fill_alpha = Number(1.0)

        self._append_fill_alpha_attr_linking_setting()

    def _append_fill_alpha_attr_linking_setting(self) -> None:
        """
        Append a scale-y attribute linking setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_fill_alpha_attr_linking_setting,
                locals_=locals(),
                module_name=__name__, class_=FillAlphaInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._fill_alpha, attr_name='fill_alpha')
            self._append_attr_to_linking_stack(
                attr=self._fill_alpha, attr_name='fill_alpha')

    @property
    def fill_alpha(self) -> Number:
        """
        Get this instance's fill opacity.

        Returns
        -------
        fill_alpha : Number
            Current fill opacity (0.0 to 1.0).

        References
        ----------
        - Graphics fill_alpha interface document
            - https://simon-ritchie.github.io/apysc/graphics_fill_alpha.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> rectangle.fill_alpha = ap.Number(0.5)
        >>> rectangle.fill_alpha
        Number(0.5)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='fill_alpha', locals_=locals(),
                module_name=__name__, class_=FillAlphaInterface):
            from apysc._type import value_util
            self._initialize_fill_alpha_if_not_initialized()
            fill_alpha: ap.Number = value_util.get_copy(
                value=self._fill_alpha)
            return fill_alpha

    @fill_alpha.setter
    def fill_alpha(
            self, value: Number) -> None:
        """
        Update this instance's fill opacity.

        Parameters
        ----------
        value : float or Number
            Fill opacity to set.

        References
        ----------
        - Graphics fill_alpha interface document
            - https://simon-ritchie.github.io/apysc/graphics_fill_alpha.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='fill_alpha', locals_=locals(),
                module_name=__name__, class_=FillAlphaInterface):
            if not isinstance(value, NumberValueInterface):
                value = ap.Number(value=value)
            self._update_fill_alpha_and_skip_appending_exp(value=value)
            self._fill_alpha._append_incremental_calc_substitution_expression()
            self._append_fill_alpha_update_expression()

            self._append_fill_alpha_attr_linking_setting()

    def _append_fill_alpha_update_expression(self) -> None:
        """
        Append fill alpha updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_fill_alpha_update_expression,
                locals_=locals(),
                module_name=__name__, class_=FillAlphaInterface):
            from apysc._type import value_util
            value_str: str = value_util.get_value_str_for_expression(
                value=self._fill_alpha)
            expression: str = (
                f'{self.variable_name}.fill({{opacity: {value_str}}});'
            )
            ap.append_js_expression(expression=expression)

    def _update_fill_alpha_and_skip_appending_exp(
            self, *, value: Number) -> None:
        """
        Update fill opacity and skip appending expression.

        Parameters
        ----------
        value : Number
            Fill opacity to set.
        """
        from apysc._converter import cast
        from apysc._validation import color_validation
        from apysc._validation import number_validation
        self._initialize_fill_alpha_if_not_initialized()
        number_validation.validate_num(num=value)
        if not isinstance(value, Number):
            value = cast.to_float_from_int(int_or_float=value)
            color_validation.validate_alpha_range(alpha=value)
            value = Number(value=value)
        color_validation.validate_alpha_range(alpha=value.value)
        self._fill_alpha = value

    _fill_alpha_snapshots: Dict[str, float]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_fill_alpha_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_fill_alpha_snapshots',
            value=self._fill_alpha._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._fill_alpha._value = self._fill_alpha_snapshots[snapshot_name]



"""Class implementation for the rotation_around_point interface.
"""

from typing import Any
from typing import Dict

from apysc._animation.animation_rotation_around_point_interface import \
    AnimationRotationAroundPointInterface
from apysc._type.dictionary import Dictionary
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class RotationAroundPointInterface(
        AnimationRotationAroundPointInterface, RevertInterface):

    _rotation_around_point: Dictionary[str, Int]

    def _initialize_rotation_around_point_if_not_initialized(self) -> None:
        """
        Initialize the `_rotation_around_point` attribute if it hasn't
        been initialized yet.
        """
        if hasattr(self, '_rotation_around_point'):
            return
        self._rotation_around_point = Dictionary({})

    def get_rotation_around_point(self, x: Int, y: Int) -> Int:
        """
        Get a rotation value around the given coordinates.

        Parameters
        ----------
        x : Int
            X-coordinate.
        y : Int
            Y-coordinate.

        Returns
        -------
        rotation : Int
            Rotation value around the given coordinates.

        References
        ----------
        - GraphicsBase rotate_around_point interfaces document
            - https://bit.ly/37TDwKs
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.get_rotation_around_point, locals_=locals(),
                module_name=__name__, class_=RotationAroundPointInterface):
            from apysc._display import rotation_interface_helper
            from apysc._type.expression_string import ExpressionString
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=x)
            number_validation.validate_integer(integer=y)
            self._initialize_rotation_around_point_if_not_initialized()
            default_val: ap.Int = ap.Int(0)
            key_exp_str: ExpressionString = rotation_interface_helper.\
                get_coordinates_key_for_expression(
                    x=int(x._value), y=int(y._value))
            rotation: ap.Int = self._rotation_around_point.get(
                key=key_exp_str, default=default_val)
            return rotation

    def set_rotation_around_point(
            self, rotation: Int, x: Int, y: Int) -> None:
        """
        Update a rotation value around the given coordinates.

        Parameters
        ----------
        rotation : Int
            Rotation value to set.
        x : Int
            X-coordinate.
        y : Int
            Y-coordinate.

        References
        ----------
        - GraphicsBase rotate_around_point interfaces document
            - https://bit.ly/37TDwKs
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.set_rotation_around_point, locals_=locals(),
                module_name=__name__, class_=RotationAroundPointInterface):
            from apysc._display import rotation_interface_helper
            from apysc._type.expression_string import ExpressionString
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=rotation)
            number_validation.validate_integer(integer=x)
            number_validation.validate_integer(integer=y)
            self._initialize_rotation_around_point_if_not_initialized()
            key_exp_str: ExpressionString = rotation_interface_helper.\
                get_coordinates_key_for_expression(
                    x=int(x._value), y=int(y._value))
            self._rotation_around_point._value[key_exp_str.value] = rotation
            self._append_rotation_around_point_update_expression(
                rotation=rotation, x=x, y=y)

    def _append_rotation_around_point_update_expression(
            self, *, rotation: Int, x: Int, y: Int) -> None:
        """
        Append a rotation value around the given coordinates
        updating expression.

        Parameters
        ----------
        rotation : Int
            Rotation value to set.
        x : Int
            X-coordinate.
        y : Int
            Y-coordinate.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_rotation_around_point_update_expression,  # noqa
                locals_=locals(),
                module_name=__name__, class_=RotationAroundPointInterface):
            expression: str = \
                self._get_rotation_around_point_updating_expression(
                    rotation=rotation, x=x, y=y)
            ap.append_js_expression(expression=expression)

    def _get_rotation_around_point_updating_expression(
            self, *, rotation: Int, x: Int, y: Int) -> str:
        """
        Get a rotation value around the given coordinates updating
        expression string.

        Parameters
        ----------
        rotation : Int
            Rotation value to set.
        x : Int
            X-coordinate.
        y : Int
            Y-coordinate.

        Returns
        -------
        expression : str
            A rotation value around the given coordinates updating
            expression string.
        """
        from apysc._display import rotation_interface_helper
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        from apysc._type import value_util
        from apysc._type.expression_string import ExpressionString
        self._initialize_rotation_around_point_if_not_initialized()
        before_value_str: str = expression_variables_util.\
            get_next_variable_name(type_name=var_names.INT)
        key_exp_str: ExpressionString = rotation_interface_helper.\
            get_coordinates_key_for_expression(x=x, y=y)
        after_value_str: str = value_util.get_value_str_for_expression(
            value=rotation)
        x_value_str: str = value_util.get_value_str_for_expression(
            value=x)
        y_value_str: str = value_util.get_value_str_for_expression(
            value=y)
        rotation_around_point_value_str: str = value_util.\
            get_value_str_for_expression(
                value=self._rotation_around_point)
        expression: str = (
            f'if ({key_exp_str.value} in '
            f'{rotation_around_point_value_str}) {{'
            f'\n  var {before_value_str} = '
            f'{rotation_around_point_value_str}[{key_exp_str.value}];'
            '\n}else {'
            f'\n  {before_value_str} = 0;'
            '\n}'
            f'\n{self.variable_name}.rotate('
            f'-{before_value_str}, {x_value_str}, {y_value_str});'
            f'\n{self.variable_name}.rotate('
            f'{after_value_str}, {x_value_str}, {y_value_str});'
            f'\n{rotation_around_point_value_str}[{key_exp_str.value}] = '
            f'{after_value_str};'
        )
        return expression

    _rotation_around_point_snapshots: Dict[str, Dict[str, Any]]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_rotation_around_point_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_rotation_around_point_snapshots',
            value={**self._rotation_around_point._value},
            snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._rotation_around_point._value = \
            self._rotation_around_point_snapshots[snapshot_name]



"""Validations related to color.
"""

from string import hexdigits
from typing import Tuple
from typing import TypeVar
from typing import Union

from apysc._type.number_value_interface import NumberValueInterface
from apysc._type.string import String

StrOrString = TypeVar('StrOrString', str, String)


def validate_hex_color_code_format(hex_color_code: StrOrString) -> None:
    """
    Validate a specified hexadecimal color code's format.

    Parameters
    ----------
    hex_color_code : str
        Hexadecimal color code (not including '#').
        e.g., 'ff0000', '666', '0'

    Raises
    ------
    ValueError
        If invalid hex color code specified.
    """
    if not isinstance(hex_color_code, (str, String)):
        raise ValueError(
            'Hex color code only supports str type, specified: '
            f'{type(hex_color_code)}')

    if isinstance(hex_color_code, String):
        value_: str = hex_color_code._value
    else:
        value_ = hex_color_code

    char_len: int = len(value_)
    expected_char_lengths: Tuple[int, int, int] = (1, 3, 6)
    if char_len not in expected_char_lengths:
        raise ValueError(
            'Not supported hex color code number of digits is specified.'
            f'\nSupported number of digits are: {expected_char_lengths}'
            f'\nSpecified: {hex_color_code} ({char_len} digits)')

    for char in value_:
        if char in hexdigits:
            continue
        raise ValueError(
            'Invalid hexadecimal character is specified.'
            f'\nTarget character: {char}'
            f'\nSupported characters: {hexdigits}')


def validate_alpha_range(
        alpha: Union[float, NumberValueInterface]) -> None:
    """
    Validate specified alpha (opacity) value's range.

    Parameters
    ----------
    alpha : float or Number
        Opacity value to check.

    Raises
    ------
    ValueError
        If specified opacity is out of 0.0 to 1.0 range.
    """
    if alpha < 0.0:
        raise ValueError(
            f'Can\'t specify alpha value less than 0.0: {alpha}')
    if alpha > 1.0:
        raise ValueError(
            f'Can\'t specify alpha value greater than 1.0: {alpha}')



"""Class implementation for number value interface.
"""

from typing import Any
from typing import Dict
from typing import Generic
from typing import TypeVar
from typing import Union

from apysc._event.custom_event_interface import CustomEventInterface
from apysc._type.copy_interface import CopyInterface
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface

_NumType = Union[int, float, 'NumberValueInterface']
_V = TypeVar('_V', int, float)
_T = TypeVar('_T', bound='NumberValueInterface')


class NumberValueInterface(
        CopyInterface, RevertInterface, CustomEventInterface,
        Generic[_V, _T]):

    _initial_value: _NumType
    _value: _V

    def __init__(
            self, *, value: _NumType,
            type_name: str) -> None:
        """
        Class for number value interface.

        Parameters
        ----------
        value : int or float or NumberValueInterface
            Initial number value.
        type_name : str
            This instance expression's type name (e.g., int, number).
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._validation import number_validation
            number_validation.validate_num(
                num=value)  # type: ignore
            self._initial_value = value
            if isinstance(value, NumberValueInterface):
                value_: _V = value._value
            else:
                value_ = value  # type: ignore
            self._value = value_
            self._type_name = type_name

    def append_constructor_expression(self) -> None:
        """
        Append current value's constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(self._initial_value, NumberValueInterface):
                value_: Union[int, float, str] = \
                    self._initial_value.variable_name
            else:
                value_ = self._value
            expression: str = (
                f'var {self.variable_name} = {value_};'
            )
            ap.append_js_expression(expression=expression)

    @property
    def value(self) -> _NumType:
        """
        Get a current number value.

        Returns
        -------
        value : int or float
            Current number value.

        References
        ----------
        - apysc basic data classes common value interface
            - https://bit.ly/3Be1aij

        Examples
        --------
        >>> import apysc as ap
        >>> int_val: ap.Int = ap.Int(10)
        >>> int_val.value
        10

        >>> int_val.value = 20
        >>> int_val.value
        20

        >>> int_val.value = ap.Int(30)
        >>> int_val.value
        30
        """
        return self._value

    @value.setter
    def value(self, value: _NumType) -> None:
        """
        Set number value.

        Parameters
        ----------
        value : int or float or NumberValueInterface
            Any number value to set.

        References
        ----------
        apysc basic data classes common value interface
            https://bit.ly/3Be1aij
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='value', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            self._set_value_and_skip_expression_appending(value=value)
            if isinstance(value, NumberValueInterface):
                self._append_value_setter_expression(value=value)
            else:
                self._append_value_setter_expression(value=self._value)

    def _set_value_and_skip_expression_appending(
            self, *, value: _NumType) -> None:
        """
        Update value attribute and skip expression appending.

        Parameters
        ----------
        value : int or float or NumberValueInterface
            Any number value to set.
        """
        from apysc._validation import number_validation
        number_validation.validate_num(
            num=value)  # type: ignore
        if isinstance(value, NumberValueInterface):
            value_: _V = value._value
        else:
            value_ = value  # type: ignore
        self._value = value_

    def _append_value_setter_expression(
            self, *, value: _NumType) -> None:
        """
        Append value's setter expresion.

        Parameters
        ----------
        value : int or float or NumberValueInterface
            Any number value to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_value_setter_expression,
                locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(value, NumberValueInterface):
                right_value: Union[str, int, float] = value.variable_name
            else:
                right_value = value
            expression: str = (
                f'{self.variable_name} = {right_value};'
            )
            ap.append_js_expression(expression=expression)

    def __add__(self, other: _NumType) -> Any:
        """
        Method for addition.

        Parameters
        ----------
        other : int or float or NumberValueInterface
            Other value to add.

        Returns
        -------
        result : NumberValueInterface
            Addition result value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__add__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(other, NumberValueInterface):
                value: _NumType = self._value + other._value
            else:
                value = self._value + other
            result: NumberValueInterface = self._copy()
            result._set_value_and_skip_expression_appending(value=value)
            self._append_addition_expression(result=result, other=other)
            return result

    def _append_addition_expression(
            self, *, result: VariableNameInterface,
            other: _NumType) -> None:
        """
        Append addition expression.

        Parameters
        ----------
        result : NumberValueInterface
            Addition result value.
        other : int or float or NumberValueInterface
            Other value to add.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_addition_expression, locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._type.value_util import get_value_str_for_expression
            right_value: str = get_value_str_for_expression(value=other)
            expression: str = (
                f'var {result.variable_name} = '
                f'{self.variable_name} + {right_value};'
            )
            ap.append_js_expression(expression=expression)

    def __sub__(self, other: _NumType) -> Any:
        """
        Method for subtraction.

        Parameters
        ----------
        other : int or float or NumberValueInterface
            Other value to subtract.

        Returns
        -------
        result : NumberValueInterface
            Subtraction result value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__sub__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(other, NumberValueInterface):
                value: _NumType = self._value - other._value
            else:
                value = self._value - other
            result: NumberValueInterface = self._copy()
            result._set_value_and_skip_expression_appending(value=value)
            self._append_subtraction_expression(result=result, other=other)
            return result

    def _append_subtraction_expression(
            self, *, result: VariableNameInterface,
            other: _NumType) -> None:
        """
        Append subtraction expression.

        Parameters
        ----------
        result : NumberValueInterface
            Subtraction result value.
        other : int or float or NumberValueInterface
            Other value to subtract.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_subtraction_expression,
                locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._type.value_util import get_value_str_for_expression
            right_value: str = get_value_str_for_expression(value=other)
            expression: str = (
                f'var {result.variable_name} = '
                f'{self.variable_name} - {right_value};'
            )
            ap.append_js_expression(expression=expression)

    def __mul__(self, other: _NumType) -> _T:
        """
        Method for multiplication.

        Parameters
        ----------
        other : int or float or NumberValueInterface
            Other value to multiply.

        Returns
        -------
        result : NumberValueInterface
            Multiplication result value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__mul__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(other, NumberValueInterface):
                value: _NumType = self._value * other._value
            else:
                value = self._value * other
            result: _T = self._copy()
            result._set_value_and_skip_expression_appending(value=value)
            self._append_multiplication_expression(result=result, other=other)
            return result

    def _append_multiplication_expression(
            self, *, result: VariableNameInterface,
            other: _NumType) -> None:
        """
        Append multiplication expression.

        Parameters
        ----------
        result : NumberValueInterface
            Multiplication result value.
        other : int or float or NumberValueInterface
            Other value to multiply.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_multiplication_expression,
                locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._type.value_util import get_value_str_for_expression
            right_value: str = get_value_str_for_expression(value=other)
            expression: str = (
                f'var {result.variable_name} = '
                f'{self.variable_name} * {right_value};'
            )
            ap.append_js_expression(expression=expression)

    def __truediv__(self, other: _NumType) -> Any:
        """
        Method for true division (return floating point number).

        Parameters
        ----------
        other : int or float or NumberValueInterface
            Other value for true division.

        Returns
        -------
        result : Number
            True division result value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__truediv__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            result: ap.Number = ap.Number(value=self)
            if isinstance(other, NumberValueInterface):
                value: _NumType = result._value / other._value
            else:
                value = result._value / other
            result._set_value_and_skip_expression_appending(value=value)
            self._append_true_division_expression(result=result, other=other)
            return result

    def _append_true_division_expression(
            self, *, result: VariableNameInterface,
            other: _NumType) -> None:
        """
        Append true division expression.

        Parameters
        ----------
        result : NumberValueInterface
            True division result value.
        other : int or float or NumberValueInterface
            Other value for true division.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_true_division_expression,
                locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._type.value_util import get_value_str_for_expression
            right_value: str = get_value_str_for_expression(value=other)
            expression: str = (
                f'{result.variable_name} = {self.variable_name} / '
                f'{right_value};'
            )
            ap.append_js_expression(expression=expression)

    def __floordiv__(self, other: _NumType) -> Any:
        """
        Method for floor division (return integer).

        Parameters
        ----------
        other : int or float or NumberValueInterface
            Other value for floor division.

        Returns
        -------
        result : Int
            Floor division result value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__floordiv__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            result: ap.Int = ap.Int(value=self)
            if isinstance(other, NumberValueInterface):
                value: _NumType = self._value // other._value
            else:
                value = self._value // other
            result._set_value_and_skip_expression_appending(value=value)
            self._append_floor_division_expression(result=result, other=other)
            return result

    def _append_floor_division_expression(
            self, *, result: VariableNameInterface,
            other: _NumType) -> None:
        """
        Append floor division expression.

        Parameters
        ----------
        result : NumberValueInterface
            Floor division result value.
        other : int or float or NumberValueInterface
            Other value for floor division.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_floor_division_expression,
                locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._type.value_util import get_value_str_for_expression
            right_value: str = get_value_str_for_expression(value=other)
            expression: str = (
                f'{result.variable_name} = '
                f'parseInt({self.variable_name} / {right_value});'
            )
            ap.append_js_expression(expression=expression)

    _incremental_calc_prev_name: str = ''

    def _append_incremental_calc_substitution_expression(self) -> None:
        """
        Append a incremental calculation's substitution expression.
        This method will be called from the each interface.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_incremental_calc_substitution_expression,  # noqa
                locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._expression import expression_variables_util
            expression_variables_util.\
                append_substitution_expression_with_names(
                    left_variable_name=self._incremental_calc_prev_name,
                    right_variable_name=self.variable_name)
            self._incremental_calc_prev_name = ''

    def __iadd__(self, other: _NumType) -> _T:
        """
        Method for incremental addition.

        Parameters
        ----------
        other : int or float or NumberValueInterface
            Other value for` incremental addition.

        Returns
        -------
        result : NumberValueInterface
            Incremental addition result value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__iadd__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._expression import expression_variables_util
            self._incremental_calc_prev_name = \
                self._get_previous_variable_name()
            result: _T = self + other
            expression_variables_util.append_substitution_expression(
                left_value=self, right_value=result)
            result.variable_name = self.variable_name
            return result

    def __isub__(self, other: _NumType) -> Any:
        """
        Method for incremental subtraction.

        Parameters
        ----------
        other : int or float or NumberValueInterface
            Other value for incremental subtraction.

        Returns
        -------
        result : NumberValueInterface
            Incremental subtraction result value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__isub__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._expression import expression_variables_util
            self._incremental_calc_prev_name = \
                self._get_previous_variable_name()
            result: NumberValueInterface = self - other
            expression_variables_util.append_substitution_expression(
                left_value=self, right_value=result)
            result.variable_name = self.variable_name
            return result

    def __imul__(self, other: _NumType) -> _T:
        """
        Method for incremental multiplication.

        Parameters
        ----------
        other : int or float or NumberValueInterface
            Other value for incremental multiplication.

        Returns
        -------
        result : NumberValueInterface
            Incremental multiplication result value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__imul__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._expression import expression_variables_util
            self._incremental_calc_prev_name = \
                self._get_previous_variable_name()
            result: _T = self * other
            expression_variables_util.append_substitution_expression(
                left_value=self, right_value=result)
            result.variable_name = self.variable_name
            return result

    def __itruediv__(self, other: _NumType) -> Any:
        """
        Method for incremental true division.

        Parameters
        ----------
        other : int or float or NumberValueInterface
            Other value for incremental true division.

        Returns
        -------
        result : NumberValueInterface
            Incremental true division result value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__itruediv__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._expression import expression_variables_util
            self._incremental_calc_prev_name = \
                self._get_previous_variable_name()
            result: NumberValueInterface = self / other
            result._incremental_calc_prev_name = \
                self._incremental_calc_prev_name
            expression_variables_util.append_substitution_expression(
                left_value=self, right_value=result)
            result.variable_name = self.variable_name
            return result

    def __mod__(self, other: _NumType) -> _T:
        """
        Method for the modulo operation.

        Parameters
        ----------
        other : int or float or NumberValueInterface
            Other value to be used in the modulo operation.

        Returns
        -------
        result : NumberValueInterface
            Modulo operation result value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__mod__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(other, NumberValueInterface):
                value: Union[int, float] = self._value % other._value
            else:
                value = self._value % other
            result: _T = self._copy()
            result._set_value_and_skip_expression_appending(value=value)
            self._append_modulo_expression(result=result, other=other)
            return result

    def _append_modulo_expression(
            self, *, result: VariableNameInterface,
            other: _NumType) -> None:
        """
        Append a module expression.

        Parameters
        ----------
        result : NumberValueInterface
            Modulo operation result value.
        other : int or float or NumberValueInterface
            Other value to be used in the modulo operation.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_modulo_expression, locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            from apysc._type.value_util import get_value_str_for_expression
            right_value: str = get_value_str_for_expression(value=other)
            expression: str = (
                f'var {result.variable_name} = '
                f'{self.variable_name} % {right_value};'
            )
            ap.append_js_expression(expression=expression)

    def __str__(self) -> str:
        """
        String conversion method.

        Returns
        -------
        string : str
            Converted value string.
        """
        if not hasattr(self, '_value'):
            return '0'
        return str(self._value)

    def __int__(self) -> int:
        """
        Integer conversion method.

        Returns
        -------
        integer : int
            Converted integer value.
        """
        return int(self._value)

    def __float__(self) -> float:
        """
        Float conversion method.

        Returns
        -------
        float_ : float
            Converted float value.
        """
        return float(self._value)

    def __eq__(self, other: Any) -> Any:
        """
        Equal comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. Builtin types, Int,
            and Number class instances are acceptable.

        Returns
        -------
        result : Boolean
            If specified value is same amount, True will be returned.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__eq__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(other, NumberValueInterface):
                result: ap.Boolean = ap.Boolean(self._value == other._value)
            else:
                result = ap.Boolean(self._value == other)
            other = self._convert_other_val_to_int_or_number(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_eq_expression(result=result, other=other)
            return result

    def _convert_other_val_to_int_or_number(self, *, other: Any) -> Any:
        """
        If comparison other value is int or float, then
        convert it to Int or Number.

        Parameters
        ----------
        other : *
            Other comparison value.

        Returns
        -------
        converted_val : *
            Converted value. If int is specified, then this will be
            Int. float is specified, then Number.
            Other type will be returned directly (not to be converted).
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._convert_other_val_to_int_or_number,
                locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(other, int):
                return ap.Int(other)
            if isinstance(other, float):
                return ap.Number(other)
            return other

    def _append_eq_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __eq__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_eq_expression, locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} === {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __ne__(self, other: Any) -> Any:
        """
        Not equal comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. Builtin types, Int,
            and Number class instances are acceptable.

        Returns
        -------
        result : Boolean
            If specified value is not same amount, True will be returned.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__ne__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(other, NumberValueInterface):
                result: ap.Boolean = ap.Boolean(self._value != other._value)
            else:
                result = ap.Boolean(self._value != other)
            other = self._convert_other_val_to_int_or_number(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_ne_expression(result=result, other=other)
            return result

    def _append_ne_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __ne__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_ne_expression, locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} !== {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __lt__(self, other: Any) -> Any:
        """
        Less than comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. Builtin types, Int,
            and Number class instances are acceptable.

        Returns
        -------
        result : Boolean
            If this value is less than a specified value, then
            True will be returned.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__lt__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(other, NumberValueInterface):
                result: ap.Boolean = ap.Boolean(self._value < other._value)
            else:
                result = ap.Boolean(self._value < other)
            other = self._convert_other_val_to_int_or_number(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_lt_expression(result=result, other=other)
            return result

    def _append_lt_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __lt__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_lt_expression, locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} < {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __le__(self, other: Any) -> Any:
        """
        Less than equal comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. Builtin types, Int,
            and Number class instances are acceptable.

        Returns
        -------
        result : Boolean
            If this value is less than or equal to a specified value,
            then True will be returned.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__le__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(other, NumberValueInterface):
                result: ap.Boolean = ap.Boolean(self._value <= other._value)
            else:
                result = ap.Boolean(self._value <= other)
            other = self._convert_other_val_to_int_or_number(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_le_expression(result=result, other=other)
            return result

    def _append_le_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __le__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_le_expression, locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} <= {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __gt__(self, other: Any) -> Any:
        """
        Greater than comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. Builtin types, Int,
            and Number class instances are acceptable.

        Returns
        -------
        result : Boolean
            If this value is greater than a specified value, then
            True will be returned.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__gt__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(other, NumberValueInterface):
                result: ap.Boolean = ap.Boolean(self._value > other._value)
            else:
                result = ap.Boolean(self._value > other)
            other = self._convert_other_val_to_int_or_number(other=other)
            if isinstance(other, NumberValueInterface):
                self._append_gt_expression(result=result, other=other)
            return result

    def _append_gt_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __gt__ expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_gt_expression, locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} > {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __ge__(self, other: Any) -> Any:
        """
        Greater than equal comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. Builtin types, Int,
            and Number class instances are acceptable.

        Returns
        -------
        result : Boolean
            If this value is greater than or equal to a specified value,
            then True will be returned.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__ge__', locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            if isinstance(other, NumberValueInterface):
                result: ap.Boolean = ap.Boolean(self._value >= other._value)
            else:
                result = ap.Boolean(self._value >= other)
            other = self._convert_other_val_to_int_or_number(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_ge_expression(result=result, other=other)
            return result

    def _append_ge_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __ge__ expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_ge_expression, locals_=locals(),
                module_name=__name__, class_=NumberValueInterface):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} >= {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    _value_snapshots: Dict[str, _V]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name='_value_snapshots',
            value=self._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._value = self._value_snapshots[snapshot_name]



"""Each types common value utilities.

Mainly following interfaces are defined:

- get_value_str_for_expression
    Get a value string for expression.
- get_copy
    Get a copy of specified instance if it is instance of CopyInterface.
"""

from typing import Any
from typing import Dict
from typing import List
from typing import TypeVar
from typing import Union

from apysc._type.array import Array


def get_value_str_for_expression(*, value: Any) -> str:
    """
    Get a value string for expression.

    Parameters
    ----------
    value : *
        Any value to convert to string.

    Returns
    -------
    value_str : str
        String for expression. If value is instance of
        VariableNameInterface, then variable's name will be returned,
        otherwise string casted value will be returned.
        Bool value will be lowercase (true or false) and str value
        will be quoted by double quotation.
        List or tuple value will be converted to js Array expression,
        e.g., '[10, "Hello!", true, any_variable]'.
        None will be Nan.
    """
    from apysc._type.expression_string import ExpressionString
    from apysc._type.variable_name_interface import VariableNameInterface
    if isinstance(value, ExpressionString):
        return value.value
    if isinstance(value, VariableNameInterface):
        return value.variable_name
    if isinstance(value, bool):
        return str(value).lower()
    if isinstance(value, str):
        return f'"{value}"'
    if isinstance(value, (list, tuple)):
        value_str: str = _get_value_str_from_iterable(value=value)
        return value_str
    if isinstance(value, dict):
        values_str = _get_value_str_from_dict(value=value)
        return values_str
    if value is None:
        return 'null'
    return str(value)


def _get_value_str_from_dict(*, value: Dict[Any, Any]) -> str:
    """
    Get a value string from dictionary object.

    Parameters
    ----------
    value : dict
        Dictionary object to convert to string.

    Returns
    -------
    value_str : str
        Converted string, e.g., '{"any_key": 10, "other_key": any_variable}'
    """
    value_str: str = '{'
    for key, value in value.items():
        if value_str != '{':
            value_str += ', '
        _validate_dict_key_type(key=key)
        value_str += f'"{key}": {get_value_str_for_expression(value=value)}'
    value_str += '}'
    return value_str


def _validate_dict_key_type(*, key: Any) -> None:
    """
    Validate whether a dictionary key type is str or int.

    Parameters
    ----------
    key : *
        Dictionary key to validate.

    Raises
    ------
    TypeError
        If key type isn't str or int.
    """
    if isinstance(key, (str, int, float)):
        return
    raise TypeError(
        f'Dictionary key type only supports str and int: {type(key)}')


def _get_value_str_from_iterable(
        *, value: Union[list, tuple, Array]) -> str:
    """
    Get a value string from iterable object.

    Parameters
    ----------
    value : list or tuple or Array
        Target iterable object.

    Returns
    -------
    value_str : str
        Converted string, e.g., '[10, "Hello!", true, any_variable]'.
    """
    from apysc._type.variable_name_interface import VariableNameInterface
    if isinstance(value, Array):
        value_: List[Any] = value.value  # type: ignore
    elif isinstance(value, tuple):
        value_ = list(value)
    else:
        value_ = value
    value_str: str = '['
    for unit_value in value_:
        if value_str != '[':
            value_str += ', '
        if isinstance(unit_value, VariableNameInterface):
            value_str += f'{unit_value.variable_name}'
            continue
        value_str += get_value_str_for_expression(value=unit_value)
    value_str += ']'
    return value_str


T = TypeVar('T')


def get_copy(*, value: T) -> T:
    """
    Get a copy of specified instance if it is instance of CopyInterface.

    Parameters
    ----------
    value : *
        Any value to copy.

    Returns
    -------
    copied : *
        Copied value. If value is not instance of CopyInterface,
        then argument value will be returned directly.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=get_copy, locals_=locals(),
            module_name=__name__):
        from apysc._type.copy_interface import CopyInterface
        if not isinstance(value, CopyInterface):
            return value
        return value._copy()



"""`trace` (console.log expression) interface implementations
"""

from typing import Any
from typing import List


def trace(*args: Any) -> None:
    """
    Display arguments information to console. This function will
    save js `console.log` expression.

    Parameters
    ----------
    *args : list
        Any arguments to display to console.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=trace, locals_=locals(),
            module_name=__name__):
        from apysc._string import string_util
        from apysc._type.variable_name_interface import VariableNameInterface

        expression: str = 'console.log('
        arg_strs: List[str] = []
        for arg in args:
            if isinstance(arg, VariableNameInterface):
                arg_strs.append(f'{arg.variable_name}')
                continue
            arg = string_util.escape_str(string=str(arg))
            arg_strs.append(f'"{arg}"')
        expression += ', '.join(arg_strs)
        expression += ');'
        ap.append_js_expression(expression=expression)



"""Class implementation for string.
"""

from typing import Any
from typing import Dict
from typing import Union

from apysc._event.custom_event_interface import CustomEventInterface
from apysc._type.copy_interface import CopyInterface
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class String(CopyInterface, RevertInterface, CustomEventInterface):
    """
    String class for apysc library.

    References
    ----------
    - String document
        - https://simon-ritchie.github.io/apysc/string.html
    - String class comparison operations document
        - https://bit.ly/3ewROEr
    - String class addition and multiplication operations document
        - https://bit.ly/2URRhWL
    """

    _initial_value: Union[str, 'String']
    _value: str

    def __init__(self, value: Union[str, 'String']) -> None:
        """
        String class for apysc library.

        Parameters
        ----------
        value : str or String
            Initial string value.

        References
        ----------
        - String document
            - https://simon-ritchie.github.io/apysc/string.html
        - String class comparison operations document
            - https://bit.ly/3ewROEr
        - String class addition and multiplication operations document
            - https://bit.ly/2URRhWL

        Examples
        --------
        >>> import apysc as ap
        >>> string: ap.String = ap.String('Hello')
        >>> string
        String('Hello')

        >>> string += ' World!'
        >>> string
        String('Hello World!')
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=String):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._expression.event_handler_scope import \
                TemporaryNotHandlerScope
            from apysc._validation import string_validation
            with TemporaryNotHandlerScope():
                TYPE_NAME: str = var_names.STRING
                string_validation.validate_string_type(string=value)
                self._initial_value = value
                self._type_name = TYPE_NAME
                self._value = self._get_str_value(value=value)
                self.variable_name = expression_variables_util.\
                    get_next_variable_name(type_name=TYPE_NAME)
                self._append_constructor_expression()

    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=String):
            expression: str = f'var {self.variable_name} = '
            if isinstance(self._initial_value, String):
                expression += f'{self._initial_value.variable_name};'
            else:
                expression += f'"{self._value}";'
            ap.append_js_expression(expression=expression)

    def _get_str_value(self, *, value: Union[str, 'String']) -> str:
        """
        Get a (Python's) str value from specified value.

        Parameters
        ----------
        value : str or String
            Target string value.

        Returns
        -------
        value : str
            Python's builtin str value.
        """
        if isinstance(value, String):
            return value._value
        return value

    @property
    def value(self) -> Union[str, 'String']:
        """
        Get a current string value.

        Returns
        -------
        value : str
            Current string value.

        References
        ----------
        - apysc basic data classes common value interface
            - https://bit.ly/3Be1aij

        Examples
        --------
        >>> import apysc as ap
        >>> string: ap.String = ap.String('Hello')
        >>> string.value = 'World!'
        >>> string.value
        'World!'
        """
        return self._value

    @value.setter
    def value(self, value: Union[str, 'String']) -> None:
        """
        Set string value.

        Parameters
        ----------
        value : str or String
            Any string value to set.

        References
        ----------
        apysc basic data classes common value interface
            https://bit.ly/3Be1aij
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='value', locals_=locals(),
                module_name=__name__, class_=String):
            from apysc._validation import string_validation
            string_validation.validate_string_type(string=value)
            self._value = self._get_str_value(value=value)
            self._append_value_setter_expression(value=value)

    def _append_value_setter_expression(
            self, *, value: Union[str, 'String']) -> None:
        """
        Append value's setter expression.

        Parameters
        ----------
        value : str or String
            Any string value to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_value_setter_expression,
                locals_=locals(),
                module_name=__name__, class_=String):
            expression: str = f'{self.variable_name} = '
            if isinstance(value, String):
                expression += f'{value.variable_name};'
            else:
                expression += f'"{value}";'
            ap.append_js_expression(expression=expression)

    def __add__(self, other: Union[str, 'String']) -> 'String':
        """
        Method for addition (string concatenation).

        Parameters
        ----------
        other : str or String
            Other string value to concatenate.

        Returns
        -------
        result : String
            Concatenated result string.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__add__', locals_=locals(),
                module_name=__name__, class_=String):
            from apysc._validation import string_validation
            string_validation.validate_string_type(string=other)
            if isinstance(other, String):
                value: str = self._value + other._value
            else:
                value = self._value + other
            result: String = self._copy()
            result._value = value
            self._append_addition_expression(result=result, other=other)
            return result

    def _append_addition_expression(
            self, *, result: VariableNameInterface,
            other: Union[str, 'String']) -> None:
        """
        Append addition (string concatenation) expression.

        Parameters
        ----------
        result : String
            Addition result value.
        other : str or String
            Other string value to concatenate.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_addition_expression,
                locals_=locals(),
                module_name=__name__, class_=String):
            from apysc._type.value_util import get_value_str_for_expression
            right_value: str = get_value_str_for_expression(value=other)
            expression: str = (
                f'var {result.variable_name} = '
                f'{self.variable_name} + {right_value};'
            )
            ap.append_js_expression(expression=expression)

    def __mul__(self, other: Union[int, Any]) -> 'String':
        """
        Method for multiplication (string repetition).

        Parameters
        ----------
        other : int or Int
            String repetition number.

        Returns
        -------
        result : String
            Repeated result string.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__mul__', locals_=locals(),
                module_name=__name__, class_=String):
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=other)
            if isinstance(other, ap.Int):
                value: int = other.value  # type: ignore
            else:
                value = other
            result: String = self._copy()
            result._value = result._value * value
            self._append_multiplication_expression(result=result, other=other)
            return result

    def _append_multiplication_expression(
            self, *, result: VariableNameInterface,
            other: Union[int, Any]) -> None:
        """
        Append multiplication (string repetition) expression.

        Parameters
        ----------
        result : String
            Multiplication result value.
        other : int or Int
            String repetition number.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_multiplication_expression,
                locals_=locals(),
                module_name=__name__, class_=String):
            expression: str = f'var {result.variable_name} = "";'
            expression += '\nfor (var i = 0; i < '
            if isinstance(other, ap.Int):
                expression += f'{other.variable_name}'
            else:
                expression += f'{other}'
            expression += '; i++) {'
            expression += (
                f'\n  {result.variable_name} += {self.variable_name};')
            expression += '\n}'
            ap.append_js_expression(expression=expression)

    def __iadd__(self, other: Union[str, 'String']) -> Any:
        """
        Method for incremental addition (string concatenation).

        Parameters
        ----------
        other : str or String
            Other string value to concatenate.

        Returns
        -------
        result : String
            Concatenated result string.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__iadd__', locals_=locals(),
                module_name=__name__, class_=String):
            from apysc._expression import expression_variables_util
            result: String = self + other
            expression_variables_util.append_substitution_expression(
                left_value=self, right_value=result)
            result.variable_name = self.variable_name
            return result

    def __imul__(self, other: Union[int, Any]) -> Any:
        """
        Method for incremental multiplication (string repetition).

        Parameters
        ----------
        other : int or Int
            String repetition number.

        Returns
        -------
        result : String
            Repetition result string.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__imul__', locals_=locals(),
                module_name=__name__, class_=String):
            from apysc._expression import expression_variables_util
            result: String = self * other
            expression_variables_util.append_substitution_expression(
                left_value=self, right_value=result)
            result.variable_name = self.variable_name
            return result

    def __str__(self) -> str:
        """
        Method for str conversion.

        Returns
        -------
        result : str
            Python builtins str value.
        """
        if not hasattr(self, '_value'):
            return ''
        return self._value

    def __eq__(self, other: Any) -> Any:
        """
        Method for equal comparison.

        Parameters
        ----------
        other : *
            Any value to compare.

        Returns
        -------
        result : Boolean
            Comparison result. If same value of str or String
            is specified, True will be returned.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__eq__', locals_=locals(),
                module_name=__name__, class_=String):
            if isinstance(other, str):
                result: ap.Boolean = ap.Boolean(self._value == other)
            elif isinstance(other, String):
                result = ap.Boolean(self._value == other._value)
            else:
                result = ap.Boolean(False)
            other = self._convert_other_val_to_string(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_eq_expression(result=result, other=other)
            return result

    def _convert_other_val_to_string(self, *, other: Any) -> Any:
        """
        If comparison other value is string, then convert it to
        String.

        Parameters
        ----------
        other : *
            Other comparison value.

        Returns
        -------
        converted_val : *
            Converted value. If other value is string, then this
            will be String instance. Other type will be returned
            directly (not to be converted).
        """
        if isinstance(other, str):
            return String(other)
        return other

    def _append_eq_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __eq__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_eq_expression, locals_=locals(),
                module_name=__name__, class_=String):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} === {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __ne__(self, other: Any) -> Any:
        """
        Method for not equal comparison.

        Parameters
        ----------
        other : *
            Any value to compare.

        Returns
        -------
        result : Boolean
            Comparison result. If not same value of str or String
            is specified, True will be returned.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__ne__', locals_=locals(),
                module_name=__name__, class_=String):
            if isinstance(other, str):
                result: ap.Boolean = ap.Boolean(self._value != other)
            elif isinstance(other, String):
                result = ap.Boolean(self._value != other._value)
            else:
                result = ap.Boolean(True)
            other = self._convert_other_val_to_string(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_ne_expression(result=result, other=other)
            return result

    def _append_ne_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __ne__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_ne_expression, locals_=locals(),
                module_name=__name__, class_=String):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} !== {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __lt__(self, other: Union[str, Any]) -> Any:
        """
        Method for less than comparison.

        Parameters
        ----------
        other : str or String
            String value to compare.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__lt__', locals_=locals(),
                module_name=__name__, class_=String):
            from apysc._validation import string_validation
            string_validation.validate_string_type(string=other)
            value: str = self._get_str_value(value=other)
            result: ap.Boolean = ap.Boolean(self._value < value)
            other = self._convert_other_val_to_string(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_lt_expression(result=result, other=other)
            return result

    def _append_lt_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __lt__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_lt_expression, locals_=locals(),
                module_name=__name__, class_=String):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} < {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __le__(self, other: Union[str, Any]) -> Any:
        """
        Method for less than or equal comparison.

        Parameters
        ----------
        other : str or String
            String value to compare.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__le__', locals_=locals(),
                module_name=__name__, class_=String):
            from apysc._validation import string_validation
            string_validation.validate_string_type(string=other)
            value: str = self._get_str_value(value=other)
            result: ap.Boolean = ap.Boolean(self._value <= value)
            other = self._convert_other_val_to_string(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_le_expression(result=result, other=other)
            return result

    def _append_le_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __le__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_le_expression, locals_=locals(),
                module_name=__name__, class_=String):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} <= {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __gt__(self, other: Union[str, Any]) -> Any:
        """
        Method for greater than comparison.

        Parameters
        ----------
        other : str or String
            String value to compare.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__gt__', locals_=locals(),
                module_name=__name__, class_=String):
            from apysc._validation import string_validation
            string_validation.validate_string_type(string=other)
            value: str = self._get_str_value(value=other)
            result: ap.Boolean = ap.Boolean(self._value > value)
            other = self._convert_other_val_to_string(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_gt_expression(result=result, other=other)
            return result

    def _append_gt_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __gt__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_gt_expression, locals_=locals(),
                module_name=__name__, class_=String):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} > {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __ge__(self, other: Union[str, Any]) -> Any:
        """
        Method for greater than or equal comparison.

        Parameters
        ----------
        other : str or String
            String value to compare.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__ge__', locals_=locals(),
                module_name=__name__, class_=String):
            from apysc._validation import string_validation
            string_validation.validate_string_type(string=other)
            value: str = self._get_str_value(value=other)
            result: ap.Boolean = ap.Boolean(self._value >= value)
            other = self._convert_other_val_to_string(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_ge_expression(result=result, other=other)
            return result

    def _append_ge_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __ge__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : VariableNameInterface
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_ge_expression, locals_=locals(),
                module_name=__name__, class_=String):
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} >= {other.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __int__(self) -> int:
        """
        Method for integer conversion.

        Returns
        -------
        result : int
            Converted integer value.
        """
        result: int = int(self._value)
        return result

    def __float__(self) -> float:
        """
        Method for float conversion.

        Returns
        -------
        result : float
            Converted float value.
        """
        result: float = float(self._value)
        return result

    def __repr__(self) -> str:
        """
        Get a representation string of this instance.

        Returns
        -------
        repr_str : str
            Representation string of this instance.
        """
        if not hasattr(self, '_value'):
            repr_str: str = "String('')"
        else:
            repr_str = f"String('{self._value}')"
        return repr_str

    _value_snapshots: Dict[str, str]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name='_value_snapshots',
            value=self._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._value = self._value_snapshots[snapshot_name]



"""Class implementation for the mouse event.
"""

from typing import Generic
from typing import TypeVar

from apysc._event.event import Event
from apysc._event.prevent_default_interface import PreventDefaultInterface
from apysc._event.stop_propagation_interface import StopPropagationInterface
from apysc._type.int import Int
from apysc._type.variable_name_interface import VariableNameInterface

T = TypeVar('T', bound=VariableNameInterface)


class MouseEvent(
        Event[T], Generic[T], StopPropagationInterface,
        PreventDefaultInterface):
    """
    Mouse event class.

    References
    ----------
    - Common mouse event interfaces
        - https://bit.ly/3eDWY1v
    """

    def __init__(self, this: T) -> None:
        """
        Mouse event class.

        Parameters
        ----------
        this : VariableNameInterface
            Instance that listening event (e.g., Sprite).

        References
        ----------
        - Common mouse event interfaces
            - https://bit.ly/3eDWY1v
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=MouseEvent):
            from apysc._expression import var_names
            super(MouseEvent, self).__init__(
                this=this, type_name=var_names.MOUSE_EVENT)

    @property
    def stage_x(self) -> Int:
        """
        Get the x-coordinate of the stage reference.

        Returns
        -------
        x : Int
            x-coordinate.

        References
        ----------
        - Common mouse event interfaces document
            - https://simon-ritchie.github.io/apysc/mouse_event_basic.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='stage_x', locals_=locals(),
                module_name=__name__, class_=MouseEvent):
            x: ap.Int = ap.Int(0)
            self._append_stage_x_getter_expression(x=x)
            return x

    def _append_stage_x_getter_expression(self, *, x: Int) -> None:
        """
        Append stage_x getter property expression.

        Parameters
        ----------
        x : Int
            Target x-coordinate value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_stage_x_getter_expression,
                locals_=locals(),
                module_name=__name__, class_=MouseEvent):
            from apysc._display.stage import get_stage_elem_str
            expression: str = (
                f'{x.variable_name} = {self.variable_name}.pageX - '
                f'{get_stage_elem_str()}.offset().left;'
            )
            ap.append_js_expression(expression=expression)

    @property
    def stage_y(self) -> Int:
        """
        Get the y-coordinate of the stage reference.

        Returns
        -------
        y : Int
            y-coordinate.

        References
        ----------
        - Common mouse event interfaces document
            - https://simon-ritchie.github.io/apysc/mouse_event_basic.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='stage_y', locals_=locals(),
                module_name=__name__, class_=MouseEvent):
            y: ap.Int = ap.Int(0)
            self._append_stage_y_getter_expression(y=y)
            return y

    def _append_stage_y_getter_expression(self, *, y: Int) -> None:
        """
        Append stage_y getter property expression.

        Parameters
        ----------
        y : Int
            Target y-coordinate value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_stage_y_getter_expression,
                locals_=locals(),
                module_name=__name__, class_=MouseEvent):
            from apysc._display.stage import get_stage_elem_str
            expression: str = (
                f'{y.variable_name} = {self.variable_name}.pageY - '
                f'{get_stage_elem_str()}.offset().top;'
            )
            ap.append_js_expression(expression=expression)

    @property
    def local_x(self) -> Int:
        """
        Get the local x-coordinate of the event listening instance.
        For example, if a Sprite instance is clicked, this value will be
        x-coordinate from Sprite's left end position.

        Returns
        -------
        x : Int
            x-coordinate.

        References
        ----------
        - Common mouse event interfaces document
            - https://simon-ritchie.github.io/apysc/mouse_event_basic.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='local_x', locals_=locals(),
                module_name=__name__, class_=MouseEvent):
            x: ap.Int = ap.Int(0)
            self._append_local_x_getter_expression(x=x)
            return x

    def _append_local_x_getter_expression(self, *, x: Int) -> None:
        """
        Append local_x getter property expression.

        Parameters
        ----------
        x : Int
            Target x-coordinate value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_local_x_getter_expression,
                locals_=locals(),
                module_name=__name__, class_=MouseEvent):
            stage_x: ap.Int = self.stage_x
            this: T = self.this
            expression: str = (
                f'{x.variable_name} = {stage_x.variable_name} - '
                f'get_total_x({this.variable_name});'
            )
            ap.append_js_expression(expression=expression)

    @property
    def local_y(self) -> Int:
        """
        Get the local y-coordinate of the event listening instance.
        For example, if a Sprite instance is clicked, this value will be
        y-coordinate from Sprite's top end position.

        Returns
        -------
        y : Int
            y-coordinate.

        References
        ----------
        - Common mouse event interfaces document
            - https://simon-ritchie.github.io/apysc/mouse_event_basic.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='local_y', locals_=locals(),
                module_name=__name__, class_=MouseEvent):
            y: ap.Int = ap.Int(0)
            self._append_local_y_getter_expression(y=y)
            return y

    def _append_local_y_getter_expression(self, *, y: Int) -> None:
        """
        Append local_y getter property expression.

        Parameters
        ----------
        y : Int
            Target y-coordinate value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_local_y_getter_expression,
                locals_=locals(),
                module_name=__name__, class_=MouseEvent):
            stage_y: ap.Int = self.stage_y
            this: T = self.this
            expression: str = (
                f'{y.variable_name} = {stage_y.variable_name} - '
                f'get_total_y({this.variable_name});'
            )
            ap.append_js_expression(expression=expression)



"""Class implementation for the skew-y animation value.
"""

from typing import Generic
from typing import TypeVar
from typing import Union

from apysc._animation.animation_base import AnimationBase
from apysc._animation.easing import Easing
from apysc._type.int import Int
from apysc._type.variable_name_interface import VariableNameInterface

_T = TypeVar('_T', bound=VariableNameInterface)


class AnimationSkewY(AnimationBase[_T], Generic[_T]):
    """
    The animation class for a skew-y.
    """

    _skew_y: Int
    _before_skew_y: Int
    _skew_y_diff: Int

    def __init__(
            self,
            *,
            target: _T,
            skew_y: Union[int, Int],
            duration: Union[int, Int] = 3000,
            delay: Union[int, Int] = 0,
            easing: Easing = Easing.LINEAR) -> None:
        """
        The animation class for a skew-y.

        Parameters
        ----------
        target : SkewXInterface
            A target instance of the animation target
            (e.g., `Rectangle` instance).
        skew_y : int or Int
            The final skew-y of the animation.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Raises
        ------
        TypeError
            If a specified target is not a SkewYInterface instance.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=AnimationSkewY):
            from apysc._converter import to_apysc_val_from_builtin
            from apysc._display.skew_y_interface import SkewYInterface
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            variable_name: str = expression_variables_util.\
                get_next_variable_name(type_name=var_names.ANIMATION_SKEW_Y)
            target_: VariableNameInterface = target
            if isinstance(target_, SkewYInterface):
                target_._initialize_skew_y_if_not_initialized()
                self._before_skew_y = target_._skew_y
            else:
                raise TypeError(
                    'Specified `target` argument is not a SkewYInterface '
                    f'instance: {type(target_)}')
            self._skew_y = to_apysc_val_from_builtin.\
                get_copied_int_from_builtin_val(integer=skew_y)
            self._skew_y_diff = self._skew_y - self._before_skew_y
            self._set_basic_animation_settings(
                target=target,
                duration=duration,
                delay=delay,
                easing=easing)
            super(AnimationSkewY, self).__init__(variable_name=variable_name)

    def _get_animation_func_expression(self) -> str:
        """
        Get a animation function expression.

        Returns
        -------
        expression : str
            Animation function expression.
        """
        from apysc._type import value_util
        diff_str: str = value_util.get_value_str_for_expression(
            value=self._skew_y_diff)
        return f'\n  .skew(0, {diff_str});'

    def _get_complete_event_in_handler_head_expression(self) -> str:
        """
        Get an expression to be inserted into the complete event
        handler's head.

        Returns
        -------
        expression : str
            An expression to be inserted into the complete event
            handler's head.
        """
        from apysc._display.skew_y_interface import SkewYInterface
        expression: str = ''
        if isinstance(self._target, SkewYInterface):
            self._target._initialize_skew_y_if_not_initialized()
            expression = (
                f'{self._target._skew_y.variable_name} = '
                f'{self._skew_y.variable_name};'
            )
        return expression



"""Implementations of Polygon class.
"""


from apysc._display import graphics
from apysc._display.append_line_point_interface import AppendLinePointInterface
from apysc._display.line_base import LineBase
from apysc._geom.point2d import Point2D
from apysc._type.array import Array


class Polygon(LineBase, AppendLinePointInterface):
    """
    The polygon vector graphics class.

    References
    ----------
    - Graphics draw_polygon interface document
        - https://bit.ly/3wHVZUk
    """

    def __init__(
            self, *, parent: 'graphics.Graphics',
            points: Array[Point2D]) -> None:
        """
        Create a polygon vector graphic. This is similar to Polyline
        class, but unlike that, end point and start point will be
        connected.

        Parameters
        ----------
        parent : Graphics
            Graphics instance to link this graphic.
        points : Array of Point2D
            List of polygon vertex points.

        References
        ----------
        - Graphics draw_polygon interface document
            - https://bit.ly/3wHVZUk
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Polygon):
            from apysc._display.graphics import Graphics
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            parent_graphics: Graphics = parent
            variable_name: str = expression_variables_util.\
                get_next_variable_name(type_name=var_names.POLYGON)
            super(Polygon, self).__init__(
                parent=parent, x=0, y=0, variable_name=variable_name)
            self.points = points
            self._set_initial_basic_values(parent=parent)
            self._append_constructor_expression()
            self._set_line_setting_if_not_none_value_exists(
                parent_graphics=parent_graphics)
            self._set_overflow_visible_setting()

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Polygon('<variable_name>')`).
        """
        repr_str: str = f"Polygon('{self.variable_name}')"
        return repr_str

    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Polygon):
            from apysc._display.stage import get_stage_variable_name
            stage_variable_name: str = get_stage_variable_name()
            points_var_name: str
            points_expression: str
            points_var_name, points_expression = \
                self._make_2dim_points_expression()
            expression: str = (
                f'{points_expression}'
                f'\nvar {self.variable_name} = {stage_variable_name}'
                f'\n  .polygon({points_var_name})'
                '\n  .attr({'
            )
            expression = self._append_basic_vals_expression(
                expression=expression, indent_num=2)
            expression += '\n  });'
            ap.append_js_expression(expression=expression)
            self._points_var_name = points_var_name



"""HTML related implementations.

Mainly following interfaces are defined:

- remove_first_selector_symbol_char : Remove first selector
    symbol (`.` or `#`) from string.
- append_html_to_str : Add html string to another string with line
    break and specified number's indentation.
- append_indent_to_each_script_line : Append indentation spaces to
    each script lines of specified html.
- ScriptLineUtil : The class for HTML's script line utility.
- is_script_start_tag_line : Get a boolean whether the specified
    line contains script start tag (`<script ...>`).
- is_script_end_tag_line : Get a boolean whether the specified line
    contains script end tag (`</script>`).
- wrap_expression_by_script_tag : Wrap an expression string by
    script start and end tag.
"""

import re
from typing import List
from typing import Match
from typing import Optional
from typing import Tuple
from typing import TypeVar

from apysc._type.string import String
from apysc._type.variable_name_interface import VariableNameInterface

StrOrString = TypeVar('StrOrString', str, String)


def remove_first_selector_symbol_char(
        *, str_val: StrOrString) -> StrOrString:
    """
    Remove first selector symbol (`.` or `#`) from string.

    Parameters
    ----------
    str_val : str or String
        Target string value. e.g., '#container'

    Returns
    -------
    str_val : str or String
        The string that removed first selector symbol character.

    Raises
    ------
    TypeError
        If other than str or String type value is passed.
    """
    from apysc._type import value_util
    if isinstance(str_val, str):
        if str_val.startswith('.') or str_val.startswith('#'):
            str_val = str_val[1:]  # type: ignore
        return str_val

    if isinstance(str_val, String):
        str_val_: String = value_util.get_copy(
            value=str_val)
        if str_val_._value.startswith('.') or str_val_._value.startswith('#'):
            str_val_._value = str_val_._value[1:]
        _append_remove_first_selector_symbol_char_expression(
            str_val=str_val_)
        return str_val_  # type: ignore

    raise TypeError(
        'Other than str or String type value is specified: '
        f'{type(str_val)}')


def _append_remove_first_selector_symbol_char_expression(
        *, str_val: VariableNameInterface) -> None:
    """
    Append remove_first_selector_symbol_char function's
    expression.

    Parameters
    ----------
    str_val : String
        First character removed string instance.
    """
    import apysc as ap
    var_name: str = str_val.variable_name
    expression: str = (
        f'var first_char = {var_name}.slice(0, 1);'
        '\nif (first_char === "." || first_char === "#") {'
        f'\n  {var_name} = {var_name}.slice(1);'
        '\n}'
    )
    ap.append_js_expression(expression=expression)


def append_html_to_str(
        *, to_append_html: str, dest_html: str, indent_num: int) -> str:
    """
    Add html string to another string with line break and specified
    number's indentation.

    Parameters
    ----------
    to_append_html : str
        HTML string to append.
    dest_html : str
        `to_append_html` will be appended to this string.
    indent_num : int
        Indentation's number. The spaces that multiplied this
        number by 2 will be added.

    Returns
    -------
    result : str
        HTML appended string.
    """
    from apysc._string import indent_util
    result: str = dest_html
    if result != '':
        result += '\n'
    result += indent_util.make_spaces_for_html(indent_num=indent_num)
    result += to_append_html
    return result


def append_indent_to_each_script_line(*, html: str, indent_num: int) -> str:
    """
    Append indentation spaces to each script lines of specified html.

    e.g., if the html is following string, then only `console.log` line
    will be added indentation.
    <html>
    <script type="text/javascript">
    console.log('Hello!');
    </script>
    </html>

    Parameters
    ----------
    html : str
        Target html string.
    indent_num : int
        Indentation number. e.g., if specified 1, then will be added
        two spaces.

    Returns
    -------
    result_html : str
        Indentation added html string.
    """
    from apysc._string import indent_util
    each_lines: List[str] = html.splitlines()
    result_html: str = ''
    script_line_util: ScriptLineUtil = ScriptLineUtil(html=html)
    for i, line in enumerate(each_lines):
        if result_html != '':
            result_html += '\n'
        if script_line_util.is_script_line(line_number=i + 1):
            result_html += indent_util.make_spaces_for_html(
                indent_num=indent_num)
        result_html += line
    return result_html


class ScriptLineUtil:

    html: str
    script_line_ranges: List[Tuple[int, int]]

    def __init__(self, *, html: str) -> None:
        """
        The class for HTML's script line utility.

        Parameters
        ----------
        html : str
            Target HTML string.
        """
        self.html = html
        self._set_script_line_ranges()

    def _set_script_line_ranges(self) -> None:
        """
        Set each script start and end line numbers.
        """
        self.script_line_ranges = []
        each_lines: List[str] = self.html.splitlines()
        start_line_num: int = 0
        for i, line in enumerate(each_lines):
            line_number: int = i + 1
            if is_script_start_tag_line(line=line):
                start_line_num = line_number + 1
                continue
            if is_script_end_tag_line(line=line):
                end_line_num: int = line_number - 1
                self.script_line_ranges.append(
                    (start_line_num, end_line_num))

    def is_script_line(self, *, line_number: int) -> bool:
        """
        Get a boolean value whether specified line number is script line
        or not.

        Parameters
        ----------
        line_number : int
            Target line number (start at 1, not 0).

        Returns
        -------
        result : bool
            If the target line is script line, then True will be set.
        """
        for script_line_start, script_line_end in self.script_line_ranges:
            if (script_line_start <= line_number
                    and line_number <= script_line_end):
                return True
        return False


def is_script_start_tag_line(*, line: str) -> bool:
    """
    Get a boolean whether the specified line contains script start
    tag (`<script ...>`).

    Notes
    -----
    External js script tag will not be target.
    e.g., `<script type="text/javascript" src="any_script.js"></script>`

    Parameters
    ----------
    line : str
        Target line string.

    Returns
    -------
    result : bool
        If specified line contains script start tag, then True
        will be set.
    """
    match: Optional[Match] = re.search(
        pattern=r'<script ', string=line)
    if match is None:
        return False
    if 'src=' in line:
        return False
    return True


def is_script_end_tag_line(*, line: str) -> bool:
    """
    Get a boolean whether the specified line contains script end
    tag (`</script>`).

    Notes
    -----
    External js script tag will not be target.
    e.g., `<script type="text/javascript" src="any_script.js"></script>`

    Parameters
    ----------
    line : str
        Target line string.

    Returns
    -------
    result : bool
        If specified line contains script end tag, then True
        will be set.
    """
    match: Optional[Match] = re.search(
        pattern=r'</script>', string=line)
    if match is None:
        return False
    if 'src=' in line:
        return False
    return True


def wrap_expression_by_script_tag(*, expression: str) -> str:
    """
    Wrap an expression string by script start and end tag.

    Parameters
    ----------
    expression : str
        An expression to wrap.

    Returns
    -------
    expression : str
        Wrapped expression string.
    """
    from apysc._html import html_const
    expression = (
        f'{html_const.SCRIPT_START_TAG}'
        f'\n{expression}'
        f'\n{html_const.SCRIPT_END_TAG}'
    )
    return expression



"""Class implementation for each mouse event interface's base class.
"""

from typing import Callable
from typing import Dict
from typing import Optional
from typing import TypeVar

from apysc._event.handler import HandlerData
from apysc._event.mouse_event import MouseEvent
from apysc._event.mouse_event_type import MouseEventType
from apysc._type.variable_name_interface import VariableNameInterface

_O = TypeVar('_O')
_Handler = Callable[[MouseEvent, _O], None]


class MouseEventInterfaceBase:

    def _set_mouse_event_handler_data(
            self, *, handler: _Handler[_O],
            handlers_dict: Dict[str, HandlerData],
            options: Optional[_O]) -> None:
        """
        Set a handler's data to the given dictionary.

        Parameters
        ----------
        handler : _Handler
            Callable would be called when event is dispatched.
        handlers_dict : dict
            Dictionary to be set handler's data.
        options : dict or None
            Optional arguments dictionary to be passed to handler.
        """
        from apysc._event.handler import get_handler_name
        from apysc._validation.handler_options_validation import \
            validate_options_type
        validate_options_type(options=options)
        name: str = get_handler_name(handler=handler, instance=self)
        if options is None:
            options = {}  # type: ignore
        handlers_dict[name] = {  # type: ignore
            'handler': handler,  # type: ignore
            'options': options,
        }

    def _unbind_mouse_event(
            self, *, handler: _Handler[_O], mouse_event_type: MouseEventType,
            handlers_dict: Dict[str, HandlerData]) -> None:
        """
        Unbind specified handler's mouse event.

        Parameters
        ----------
        handler : _Handler
            Callable to be unbinded.
        mouse_event_type : MouseEventType
            Event type to unbind.
        handlers_dict : dict
            Dictionary that has handler's data.
        """
        import apysc as ap
        from apysc._validation.variable_name_validation import \
            validate_variable_name_interface_type
        with ap.DebugInfo(
                callable_=self._unbind_mouse_event, locals_=locals(),
                module_name=__name__, class_=MouseEventInterfaceBase):
            from apysc._event.handler import append_unbinding_expression
            from apysc._event.handler import get_handler_name
            self_instance: VariableNameInterface = \
                validate_variable_name_interface_type(instance=self)
            name: str = get_handler_name(handler=handler, instance=self)
            if name in handlers_dict:
                del handlers_dict[name]
            append_unbinding_expression(
                this=self_instance, handler_name=name,
                mouse_event_type=mouse_event_type)

    def _unbind_all_mouse_events(
            self, *, mouse_event_type: MouseEventType,
            handlers_dict: Dict[str, HandlerData]) -> None:
        """
        Unbind specified all mouse event type's event.

        Parameters
        ----------
        mouse_event_type : MouseEventType
            Event type to unbind.
        handlers_dict : dict
            Dictionary that has handler's data.
        """
        import apysc as ap
        from apysc._validation.variable_name_validation import \
            validate_variable_name_interface_type
        with ap.DebugInfo(
                callable_=self._unbind_all_mouse_events, locals_=locals(),
                module_name=__name__, class_=MouseEventInterfaceBase):
            from apysc._event.handler import append_unbinding_all_expression
            self_instance: VariableNameInterface = \
                validate_variable_name_interface_type(instance=self)
            handlers_dict.clear()
            append_unbinding_all_expression(
                this=self_instance, mouse_event_type=mouse_event_type)

    def _append_mouse_event_binding_expression(
            self, *, name: str, mouse_event_type: MouseEventType) -> None:
        """
        Append a mouse event binding expression.

        Parameters
        ----------
        name : str
            Handler's name.
        mouse_event_type : MouseEventType
            Event type to bind.
        """
        import apysc as ap
        from apysc._validation.variable_name_validation import \
            validate_variable_name_interface_type
        with ap.DebugInfo(
                callable_=self._append_mouse_event_binding_expression,
                locals_=locals(),
                module_name=__name__, class_=MouseEventInterfaceBase):
            from apysc._type.variable_name_interface import \
                VariableNameInterface
            self_instance: VariableNameInterface = \
                validate_variable_name_interface_type(instance=self)
            expression: str = (
                f'{self_instance.variable_name}'
                f'.{mouse_event_type.value}({name});'
            )
            ap.append_js_expression(expression=expression)



"""Class implementation for the x-coordinate interface.
"""

from typing import Dict

from apysc._animation.animation_move_interface import AnimationMoveInterface
from apysc._animation.animation_x_interface import AnimationXInterface
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class XInterface(
        AnimationXInterface, AnimationMoveInterface, RevertInterface,
        AttrLinkingInterface):

    _x: Int

    def _initialize_x_if_not_initialized(self) -> None:
        """
        Initialize the _x attribute if it hasn't been initialized yet.
        """
        if hasattr(self, '_x'):
            return
        self._x = Int(0)

        self._append_x_attr_linking_setting()

    def _append_x_attr_linking_setting(self) -> None:
        """
        Append a x attribute linking setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_x_attr_linking_setting,
                locals_=locals(),
                module_name=__name__, class_=XInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._x, attr_name='x')
            self._append_attr_to_linking_stack(
                attr=self._x, attr_name='x')

    @property
    def x(self) -> Int:
        """
        Get a x-coordinate.

        Returns
        -------
        x : Int
            X-coordinate.

        References
        ----------
        - Display object x and y interfaces document
            - https://bit.ly/2ToA5ba
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='x', locals_=locals(),
                module_name=__name__, class_=XInterface):
            from apysc._type import value_util
            self._initialize_x_if_not_initialized()
            x: ap.Int = value_util.get_copy(value=self._x)
            return x

    @x.setter
    def x(self, value: Int) -> None:
        """
        Update x-coordinate.

        Parameters
        ----------
        value : int or Int
            X-coordinate value.

        References
        ----------
        - Display object x and y interfaces document
            - https://bit.ly/2ToA5ba
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='x', locals_=locals(),
                module_name=__name__, class_=XInterface):
            from apysc._type.number_value_interface import NumberValueInterface
            from apysc._validation import number_validation
            if not isinstance(value, NumberValueInterface):
                number_validation.validate_integer(integer=value)
                value = ap.Int(value=value)
            self._x = value
            self._x._append_incremental_calc_substitution_expression()
            self._append_x_update_expression()

            self._append_x_attr_linking_setting()

    def _append_x_update_expression(self) -> None:
        """
        Append the x position updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_x_update_expression, locals_=locals(),
                module_name=__name__, class_=XInterface):
            from apysc._type import value_util
            self._initialize_x_if_not_initialized()
            value_str: str = value_util.get_value_str_for_expression(
                value=self._x)
            expression: str = (
                f'{self.variable_name}.x({value_str});'
            )
            ap.append_js_expression(expression=expression)

    _x_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_x_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_x_snapshots',
            value=int(self._x._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._x._value = self._x_snapshots[snapshot_name]



"""Class implementation for copy interface.
"""

from copy import deepcopy
from typing import Any

from apysc._type.type_name_interface import TypeNameInterface
from apysc._type.variable_name_interface import VariableNameInterface


class CopyInterface(TypeNameInterface, VariableNameInterface):

    def _copy(self) -> Any:
        """
        Make a deep copy of this instance.

        Returns
        -------
        result : *
            Copied instance.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._copy, locals_=locals(),
                module_name=__name__, class_=CopyInterface):
            from apysc._expression import expression_variables_util
            from apysc._expression.event_handler_scope import \
                TemporaryNotHandlerScope
            with TemporaryNotHandlerScope():
                result: CopyInterface = deepcopy(self)
                result.variable_name = \
                    expression_variables_util.get_next_variable_name(
                        type_name=self.type_name)
                self._append_copy_expression(
                    result_variable_name=result.variable_name)
            self._append_value_updating_cpy_exp_to_handler_scope(
                result_variable_name=result.variable_name)
            return result

    def _append_value_updating_cpy_exp_to_handler_scope(
            self, *, result_variable_name: str) -> None:
        """
        Append a value updating copy expression
        if the current scope is an event handler's one.

        Parameters
        ----------
        result_variable_name : str
            Copied value's variable name.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_value_updating_cpy_exp_to_handler_scope,
                locals_=locals(),
                module_name=__name__, class_=CopyInterface):
            from apysc._expression import event_handler_scope
            from apysc._type.type_util import is_immutable_type
            evt_handler_scope_count: int = event_handler_scope.\
                get_current_event_handler_scope_count()
            if evt_handler_scope_count == 0:
                return
            if is_immutable_type(value=self):
                expression: str = (
                    f'{result_variable_name} = {self.variable_name};'
                )
            else:
                expression = (
                    f'{result_variable_name} = '
                    f'cpy({self.variable_name});'
                )
            ap.append_js_expression(expression=expression)

    def _append_copy_expression(self, *, result_variable_name: str) -> None:
        """
        Append copy expression.

        Parameters
        ----------
        result_variable_name : str
            Copied value's variable name.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_copy_expression, locals_=locals(),
                module_name=__name__, class_=CopyInterface):
            from apysc._type.type_util import is_immutable_type
            if is_immutable_type(value=self):
                expression: str = (
                    f'var {result_variable_name} = '
                    f'{self.variable_name};'
                )
            else:
                expression = (
                    f'var {result_variable_name} = '
                    f'cpy({self.variable_name});'
                )
            ap.append_js_expression(expression=expression)
